# Copyright 2022 Cognite AS
from .pump_parameters import (
    percent_BEP_flowrate,
    pump_hydraulic_power,
    pump_shaft_power,
    recycle_valve_power_loss,
    total_head,
)
from .valve_parameters import flow_through_valve


TOOLBOX_NAME = "Equipment"

__all__ = [
    "total_head",
    "percent_BEP_flowrate",
    "pump_hydraulic_power",
    "pump_shaft_power",
    "recycle_valve_power_loss",
    "flow_through_valve",
]



# Copyright 2022 Cognite AS
from collections import defaultdict
from typing import Any, Callable, DefaultDict, Dict, List, Optional, Sequence

from packaging.version import Version

from .type_check import check_types


_registered_funcs: DefaultDict[str, Dict[str, Any]] = defaultdict(dict)


@check_types
def register(
    version: str,
    name: Optional[str] = None,
    deprecated: Optional[bool] = False,
    changelog: Optional[str] = None,
) -> Callable[..., Any]:
    """Decorator to register a new versioned function.

    Args:
        version: Version number of the function
            The version number follows the standard version scheme for Python packages
            as defined in `PEP 440 <https://packaging.pypa.io/en/latest/version.html>`_.

            The following are valid version numbers (shown in the order that
            would be obtained by sorting according to the supplied cmp function):

                0.4       0.4.0  (these two are equivalent)
                0.4.1
                0.5a1
                0.5b3
                0.5
                0.9.6
                1.0
                1.0.4a3
                1.0.4b1
                1.0.4

            The following are examples of invalid version numbers:

                1.3pl1
                a
        name: Name of the function
            The name under which the function should be registered. If no name
            is specified, the name of the function is used.
        deprecated: Flag if function is deprecated. Default: False
        changelog: Description of what has changed since the previous version. Default: None
    """
    # Test that the version numbering is valid
    Version(version)

    def register_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Register a new versioned function."""
        key = name or func.__name__

        if version in _registered_funcs[key]:
            raise ValueError(f"Function {key} with version {version} is already registered")

        _registered_funcs[key][version] = func

        func.__versioning_version__ = version  # type: ignore
        func.__versioning_name__ = key  # type: ignore
        func.__versioning_deprecated__ = deprecated  # type: ignore
        func.__versioning_changelog__ = changelog  # type: ignore

        return func

    return register_decorator


def get_registered_functions() -> List[str]:
    """Return the list of registered function names."""
    return list(_registered_funcs.keys())


@check_types
def get_versions(name: str) -> List[str]:
    """Return the list of available versions of a function name.

    Sorted from low to highest version
    """
    versions = _registered_funcs[name].keys()
    return sorted(versions, key=Version)


@check_types
def get_name(func: Callable[..., Any]) -> str:
    """Return the name with which the function was registered."""
    try:
        return func.__versioning_name__  # type: ignore
    except AttributeError:
        raise ValueError("Cannot get name of an un-registerd function")


@check_types
def get_version(func: Callable[..., Any]) -> Optional[str]:
    """Return the version of a function.

    If the function is not registered, None is returned
    """
    try:
        return func.__versioning_version__  # type: ignore
    except AttributeError:
        return None


@check_types
def is_versioned(func: Callable[..., Any]) -> bool:
    """Return true if the version is versioned."""
    return hasattr(func, "__versioning_version__")


@check_types
def is_deprecated(func: Callable[..., Any]) -> Optional[bool]:
    """Return true if the version is deprecated.

    If the function is not registered, None is returned
    """
    try:
        return func.__versioning_deprecated__  # type: ignore
    except AttributeError:
        return None


@check_types
def get_changelog(func: Callable[..., Any]) -> Optional[str]:
    """Return changelog of this version.

    If the function is not registered, None is returned
    """
    try:
        return func.__versioning_changelog__  # type: ignore
    except AttributeError:
        return None


@check_types
def get(name: str, version: Optional[str] = None) -> Callable[..., Any]:
    """Return one of the versions of a function.

    If version is None, the latest version is returned
    """
    if version is None:
        version = get_versions(name)[-1]

    if name not in _registered_funcs:
        raise ValueError(f"Function {name} is not registered")
    if version not in _registered_funcs[name]:
        raise ValueError(f"Version {version} for function {name} is not registered")
    return _registered_funcs[name][version]


@check_types
def run(
    name: str, version: Optional[str] = None, args: Optional[Sequence] = None, kwargs: Optional[dict] = None
) -> Any:
    """Run a version of a function.

    If version is None, the latest version is executed
    """
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}
    return get(name, version)(*args, **kwargs)



# Copyright 2022 Cognite AS
from .change_point_detector import cpd_ed_pelt
from .cusum import cusum
from .drift_detector import drift
from .steady_state import ssd_cpd, ssid, vma
from .unchanged_signal_detector import unchanged_signal_detector


TOOLBOX_NAME = "Detect"

__all__ = ["drift", "ssid", "vma", "ssd_cpd", "cpd_ed_pelt", "unchanged_signal_detector", "cusum"]



# Copyright 2021 Cognite AS
import numpy as np
import pandas as pd

from ..exceptions import UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_time_index


@check_types
def resample_timeseries(data: pd.Series) -> pd.Series:
    """Resample timeseries.

    Resamples a timeseries (provided as a pandas Series with datetime index)
    into a equally spaced timeseries with the frequency defined by the smallest
    delta time between each timestamp.

    Args:
        data: Time series

    Returns:
        pandas.Series: Resampled Time series
    """
    validate_series_has_time_index(data)
    if len(data) < 2:
        raise UserValueError(f"Expected data to be of length >= 2, got length {len(data)}")
    # compute delta time in seconds
    delta_time: np.ndarray = np.diff(data.index.values, n=1, axis=-1).astype(np.int64) * 1e-9
    # prevent timesteps smaller than 60s
    min_delta_time: float = max(delta_time.min().round(), 60)
    # define frequency
    frequency: pd.DateOffset = pd.DateOffset(seconds=min_delta_time)

    # compute min and max time
    min_time: pd.Timestamp = data.index.min()
    max_time: pd.Timestamp = data.index.max()

    # create a new series with equally spaced timestamps and join it with input data
    index: pd.DatetimeIndex = pd.date_range(start=min_time, end=max_time, freq=frequency)
    data_equally_spaced: pd.Series = pd.DataFrame(index=index)
    data_missing_values: pd.Series = data_equally_spaced.join(data.rename("new"), how="outer").iloc[:, 0]

    # use linear interpolation to estimate missing values
    data_interpolated: pd.Series = data_missing_values.interpolate(method="linear")

    # only keep data for the equally spaced timestamps
    data_clean: pd.Series = data_equally_spaced.join(data_interpolated.rename("new"), how="left").iloc[:, 0]

    return data_clean


@check_types
def constrain(value: float, min: float = 1.0e-4, max: float = 1.0e6) -> float:
    """Constrains a value to not exceed a maximum and minimum value.

    Args:
        value: The value to constrain
        min: The minimum limit. Defaults to 1.0e-4.
        max: The maximum limit. Defaults to 1.0e6.

    Returns:
        float: Value within the specified limits
    """
    if value > max:
        return max
    elif value < min:
        return min
    else:
        return value



# Copyright 2022 Cognite AS
from typing import Literal

import pandas as pd

from scipy.signal import butter, sosfilt

from indsl import versioning
from indsl.type_check import check_types

from . import butterworth_v1  # noqa


# noinspection SpellCheckingInspection
@check_types
@versioning.register(
    version="2.0",
    changelog="Unused or irrelevant lines of code removed and filter output parameter "
    "removed from function signature and set to `sos`.",
)
def butterworth(
    data: pd.Series,
    N: int = 50,
    Wn: float = 0.1,
    btype: Literal["lowpass", "highpass"] = "lowpass",
) -> pd.Series:
    """Butterworth.

    This signal processing filter is designed to have a frequency response as flat as possible in the passband and
    roll-offs towards zero in the stopband. In other words, this filter is designed not to modify much the signal at the
    in the passband and attenuate as much as possible the signal at the stopband. At the moment, only low and high pass
    filtering are supported.

    Args:
        data: Time series.
        N: Order.
            Defaults to 50.
        Wn: Critical frequency.
            Number between 0 and 1, with 1 representing one-half of the sampling rate (Nyquist frequency).
            Defaults to 0.1.
        btype: Filter type.
            The options are: "lowpass" and "highpass"
            Defaults to "lowpass".

    Returns:
        pandas.Series: Filtered signal.
    """
    data = data.dropna()

    if len(data) < 1:
        return data

    filter_output = butter(N=N, Wn=Wn, output="sos", btype=btype)
    # Apply second order segments
    filtered = sosfilt(filter_output, data, axis=0)

    return pd.Series(filtered, index=data.index)



# Copyright 2021 Cognite AS
import numba
import numpy as np
import pandas as pd

from scipy import stats

from indsl.detect.change_point_detector import ed_pelt
from indsl.detect.utils import constrain, resample_timeseries

from ..exceptions import UserValueError

# Detectors
from ..type_check import check_types
from ..validations import validate_series_has_time_index


@check_types
def ssd_cpd(
    data: pd.Series, min_distance: int = 15, var_threshold: float = 2.0, slope_threshold: float = -3.0
) -> pd.Series:
    """Steady State Detection (CPD).

    Detect steady state periods in a time series based on a change point detection algorithm.  The time series is split
    into "statistically homogeneous" segments using the ED Pelt change point detection algorithm. Then each segment is tested with regard
    to a normalized standard deviation and the slope of the line of best fit to determine if the segment can be
    considered a steady or transient region.

    Args:
        data: Time series.

        min_distance: Minimum distance.
            Specifies the minimum point-wise distance for each segment that will be considered in the Change
            Point Detection algorithm.

        var_threshold: Variance threshold.
            Specifies the variance threshold. If the normalized variance calculated for a given segment is greater than
            the threshold, the segment will be labeled as transient (value = 0).

        slope_threshold: Slope threshold.
            Specifies the slope threshold. If the slope of a line fitted to the data of a given segment is greater than
            10 to the power of the threshold value, the segment will be labeled as transient (value = 0).

    Returns:
        pandas.Series: Binary time series.
        Steady state = 1, Transient = 0.
    """
    validate_series_has_time_index(data)

    # resample the provided timeseries into an equally spaced one
    data_resampled: pd.Series = resample_timeseries(data=data)

    # store the arrays with datapoints and timestamp (in ms since epoch)
    y: np.ndarray = data_resampled.values
    x: np.ndarray = data_resampled.index.to_numpy("datetime64[ms]").view(np.int64)

    # the maximum allowable distance is half the number of datapoints
    max_distance: int = int(np.floor(len(x) / 2))
    if min_distance > max_distance:
        raise UserValueError(
            f"min_distance ({min_distance}) is larger than the maximum allowable distance ({max_distance})"
        )

    # compute the change points
    change_points: np.ndarray = ed_pelt(data=y, min_distance=min_distance)

    # include the endpoints to slice the segments
    change_points = np.insert(change_points, 0, 0)  # index of the first datapoint
    change_points = np.insert(change_points, len(change_points), len(y))  # index of the last datapoint

    # compute mean of the data
    avg: float = np.mean(y)
    # constrains the mean of the data into predefined limits
    # this will prevent generating infinite values on the var calculation below
    divisor: float = constrain(value=avg, min=1.0e-4, max=1.0e6)

    # loop over all changepoint segments
    # we consider a region as transient (value = 0) unless it passes the subsequent logical tests
    ss_map: np.ndarray = np.zeros_like(y)
    for i in range(1, len(change_points)):
        i0: int = change_points[i - 1]
        i1: int = change_points[i]
        xi: np.ndarray = x[i0:i1]
        yi: np.ndarray = y[i0:i1]

        # compute the segment standard deviation and normalize it by the segment length
        std: float = np.std(yi) / (i1 - i0)
        # calculate a normalized variance (this is an effort to make the var_threshold more generic between different
        # timeseries)
        var: float = 1.0e5 * std / divisor

        # we first check if the variance criteria is met
        if abs(var) < var_threshold:
            # fit a line to the segment data
            lr = stats.linregress(xi, yi)
            slope = lr.slope

            # then we check if the slope criteria is met
            if abs(slope) < np.power(10.0, slope_threshold):
                ss_map[i0:i1] = 1.0

    # we finally create the series that is returned by the function
    ss_map_series: pd.Series = pd.Series(data=ss_map, index=data_resampled.index)

    return ss_map_series


@check_types
def ssid(
    data: pd.Series, ratio_lim: float = 2.5, alpha1: float = 0.2, alpha2: float = 0.1, alpha3: float = 0.1
) -> pd.Series:
    """Steady state (variance).

    The steady state detector is based on the ration of two variances estimated from the same signal [#]_ . The algorithm first
    filters the data using the factor "Alpha 1" and calculates two variances (long and short term) based on the
    parameters "Alpa 2" and "Alpha 3". The first variance is an exponentially weighted moving variance based on the
    difference between the data and the average. The second is also an exponentially weighted moving “variance” but
    based on sequential data differences. Larger Alpha values imply that fewer data are involved in the analysis,
    which has the benefit of reducing the time for the identifier to detect a process change (average run length, ARL)
    but has an undesired impact of increasing the variability on the results, broadening the distribution and
    confounding interpretation. Lower λ values undesirably increase the average run length to detection but increase
    precision (minimizing Type-I and Type-II statistical errors) by reducing the variability of the distributions
    and increasing the signal-to-noise ratio of a TS to SS situation.

    Args:
        data: Time series.
        ratio_lim: Threshold.
            Specifies the variance ratio threshold if it is in steady state or not. A variance ratio greater than the
            threshold labels the state as transient.
        alpha1: Alpha 1.
            Filter factor for the mean. Value should be between 0 and 1. Recommended value is 0.2.
            Defaults to 0.2.
        alpha2: Alpha 2.
            Filter factor for variance 1. Value should be between 0 and 1. Recommended value is 0.1.
            Defaults to 0.1.
        alpha3: Alpha 3.
            Filter factor for variance 2. Value should be between 0 and 1. Recommended value is 0.1.
            Defaults to 0.1.

    Returns:
        pandas.Series: Binary time series.
        Steady state = 0, transient = 1.

    References:
        .. [#] Rhinehart, R. Russell. (2013). Automated steady and transient state identification in noisy processes.
               Proceedings of the American Control Conference. 4477-4493. 10.1109/ACC.2013.6580530
    """
    # Determine number of datapoints to be used to initialize values (must have at least 10 within range)
    Ls = int(min(len(data) * 0.05, 50))

    if Ls < 10:
        raise UserValueError("There are too few datapoints to detect steady states.")

    # Check the bounds of alpha parameters
    for par, par_val in zip([alpha1, alpha2, alpha3], ["alpha1", "alpha2", "alpha3"]):
        if not (0 <= par <= 1):
            raise UserValueError(f"Alpha parameters should be between 0 and 1. Received {par_val} for {par}")

    # Split data into first sample points and rest of the data
    data_ini = data[0:Ls]
    df = data[Ls + 1 :]

    # Initialize filter and variance
    filt0 = data_ini[0:Ls].mean()
    var0 = data_ini[0:Ls].var()
    initial = data_ini[Ls - 1]

    start_val = df[0]
    filt = np.array([alpha1 * start_val + (1 - alpha1) * filt0])
    var1 = np.array([alpha2 * (start_val - filt0) ** 2 + (1 - alpha2) * var0])
    var2 = np.array([alpha3 * (start_val - initial) ** 2 + (1 - alpha3) * var0])

    # Calculate filter
    values = df.tolist()
    for i in range(1, len(values)):
        filt = np.append(filt, alpha1 * values[i] + (1 - alpha1) * filt[i - 1])
        var1 = np.append(var1, alpha2 * (values[i] - filt[i - 1]) ** 2 + (1 - alpha2) * var1[i - 1])
        var2 = np.append(var2, alpha3 * (values[i] - values[i - 1]) ** 2 + (1 - alpha3) * var2[i - 1])

    # Calculate Variance Ratio and find unsteady regions
    ratio = (2 - alpha1) * (var1 / var2)
    unsteady_flag = ratio > ratio_lim

    return pd.Series(unsteady_flag, index=df.index).astype(int)


@check_types
def vma(series: pd.Series, window_length: int = 10) -> pd.Series:
    """Steady state (vma).

    This moving average is designed to become flat (constant value) when the data
    within the lookup window does not vary significantly. It can also be state detector. The calculation is based on
    the variability of the signal in a lookup window.

    Args:
        series: Time series.
        window_length: Lookup window.
            Window length in data points used to estimate the variability of the signal.

    Returns:
        pandas.Series: Moving average.
        If the result has the same value as the previous moving average result, the signal can be considered to
        be on steady state.
    """
    k = 1.0 / window_length
    N = len(series)

    # Estimate iS:
    diff = np.diff(series)
    iS = _estimate_iS(diff, k, N)

    # Estimate hhv, llv, vI:
    iS_df = pd.Series(iS, index=series.index).rolling(window_length, min_periods=1).aggregate(["min", "max"]).fillna(0)
    hhv, llv = iS_df["max"], iS_df["min"]
    vI = (iS - llv) / (hhv - llv)

    # Finally, estimate variable moving average:
    return pd.Series(_estimate_vma(vI.values, series.values, k, N), index=series.index)


@numba.njit
def _estimate_iS(diff: np.ndarray, k: float, N: int) -> np.ndarray:
    # Estimate pdm and mdm:
    pdm, mdm = np.zeros(N), np.zeros(N)
    pdm[1:] = np.maximum(diff, 0)
    mdm[1:] = np.maximum(-diff, 0)
    # Estimate pdmS, mdmS:
    pdmS, mdmS = np.zeros(N), np.zeros(N)
    for i in range(1, N):
        pdmS[i] = (1 - k) * pdmS[i - 1] + k * pdm[i]
        mdmS[i] = (1 - k) * mdmS[i - 1] + k * mdm[i]
    pdi, mdi = np.zeros(N), np.zeros(N)
    s = mdmS + pdmS
    pdi[1:] = pdmS[1:] / s[1:]
    mdi[1:] = mdmS[1:] / s[1:]
    # Estimate pdiS, mdiS, d, s1:
    pdiS, mdiS = np.zeros(N), np.zeros(N)
    for i in range(1, N):
        pdiS[i] = (1 - k) * pdiS[i - 1] + k * pdi[i]
        mdiS[i] = (1 - k) * mdiS[i - 1] + k * mdi[i]
    d = np.abs(pdiS - mdiS)
    s1 = pdiS + mdiS
    # Estimate iS:
    iS = np.zeros(N)
    for i in range(1, N):
        iS[i] = (1 - k) * iS[i - 1] + k * d[i] / s1[i]
    return iS


@numba.njit
def _estimate_vma(vI: np.ndarray, arr: np.ndarray, k: float, N: int):
    vma = np.zeros(N)
    for i in range(1, N):
        vma[i] = (1 - k * vI[i]) * vma[i - 1] + k * vI[i] * arr[i]
    return vma



# Copyright 2022 Cognite AS

from typing import Optional

import numpy as np
import pandas as pd

from csaps import csaps
from kneed import KneeLocator
from sklearn.cluster import DBSCAN
from sklearn.neighbors import NearestNeighbors

from indsl import versioning
from indsl.exceptions import UserValueError
from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index, validate_series_is_not_empty

from . import outliers_v1  # noqa


@versioning.register(
    version="2.0",
    changelog="handles steady_state inputs without crashing, improves performance and implements new input types",
)
@check_types
def detect_outliers(
    data: pd.Series,
    reg_smooth: float = 0.9,
    min_samples: int = 4,
    eps: Optional[float] = None,
    time_window: pd.Timedelta = pd.Timedelta("60min"),
    del_zero_val: bool = False,
) -> pd.Series:
    """Outlier detection.

    Identifies outliers combining two methods, dbscan and csap.

    - **dbscan**: Density-based clustering algorithm used to identify clusters of varying shape and size within a data
      set. Does not require a pre-determined set number of clusters. Able to identify outliers as noise, instead of
      classifying them into a cluster. Flexible when it comes to the size and shape of clusters, which makes it more
      useful for noise, real life data.

    - **csaps regression**: Cubic smoothing spline algorithm. Residuals from the regression are computed. Data points with
      high residuals (3 Standard Deviations from the Mean) are considered as outliers.

    Args:
        data: Time series.
            The data has to be non-uniform.
        reg_smooth: Smoothing factor.
            The smoothing parameter that determines the weighted sum of terms in the regression, and it is limited by
            the range [0,1]. Defaults to 0.9. Ref: https://csaps.readthedocs.io/en/latest/formulation.html#definition
        min_samples: Minimum samples.
            Minimum number of data points required to form a distinct cluster. Defaults to 4.
            Defines the minimum number of data points required to form a distinct cluster. Rules of thumb for selecting
            the minimum samples value:

            * The larger the data set, the larger the value of MinPts should be.
            * If the data set is noisier, choose a larger value of MinPts Generally, MinPts should be greater than or
              equal to the dimensionality of the data set. For 2-dimensional data, use DBSCAN’s default value of 4
              (Ester et al., 1996).
            * If your data has more than 2 dimensions, choose MinPts = 2*dim, where dim= the dimensions of your data
              set (Sander et al., 1998).
        eps: Distance threshold.
            Defines the maximum distance between two samples for one to be considered as in the
            neighborhood of the other (i.e. belonging to the same cluster). This is the most important DBSCAN parameter
            to choose appropriately for your dataset and distance function. If no value is given, it is set automatically
            using Nearest Neighbors algorithm to calculate the average distance between each point and its k
            nearest neighbors, where k = min_samples (minimum samples). In ascending order on a k-distance graph, the
            optimal value for the threshold is at the point of maximum curvature (i.e. after plotting the average
            k-distances in where the graph has the greatest slope). This is not a maximum bound on the distances of
            points within a cluster.
            Defaults to None, eps value has to be > 0.0.
        time_window: Time window.
            Length of the time period to compute the rolling mean. The rolling mean and the data point value are the two features considered when calculating the distance to the furthest neighbour.
            This distance allows us to find the right epsilon when training dbscan. Defaults to '60min'.
            Accepted format: '3w', '10d', '5h', '30min', '10s'.
            If a number without unit (such as '60')is given, it will be considered as the number of minutes.
        del_zero_val: Remove zeros.
            Removes data points containing a value of 0. Defaults to False.

    Returns:
        pandas.Series: Time series.
            Binary time series indicating outliers: Outlier= 1, Not an outlier = 0
    """
    outlier_indices = _get_outlier_indices(data, min_samples, eps, time_window, del_zero_val, reg_smooth)

    df_with_outliers = pd.DataFrame(data=1, index=outlier_indices, columns=["isOutlier"])

    # Convert original time series to DataFrame
    df = data.to_frame()

    # Left join original time series and timeseries with outliers
    outlier_indicator_df = df.join(df_with_outliers)

    # Fill NaN with zeroes
    outlier_indicator_df["isOutlier"] = outlier_indicator_df["isOutlier"].fillna(0).astype(int)

    # Convert DataFrame to Series
    outlier_indicator_ts = outlier_indicator_df["isOutlier"].squeeze()

    return outlier_indicator_ts


@versioning.register(version="2.0", changelog="handle steady_state inputs, better performance; new input types")
@check_types
def remove_outliers(
    data: pd.Series,
    reg_smooth: float = 0.9,
    min_samples: int = 4,
    eps: Optional[float] = None,
    time_window: pd.Timedelta = pd.Timedelta("60min"),
    del_zero_val: bool = False,
) -> pd.Series:
    """Outlier removal.

    Identifies and removes outliers combining two methods, dbscan and csap.

    - **dbscan**: Density-based clustering algorithm used to identify clusters of varying shape and size within a data
      set. Does not require a pre-determined set number of clusters. Able to identify outliers as noise, instead of
      classifying them into a cluster. Flexible when it comes to the size and shape of clusters, which makes it more
      useful for noise, real life data.

    - **csaps regression**: Cubic smoothing spline algorithm. Residuals from the regression are computed. Data points with
      high residuals (3 Standard Deviations from the Mean) are considered as outliers.

    Args:
        data: Time series.
        reg_smooth: Smoothing factor.
            The smoothing parameter that determines the weighted sum of terms in the regression, and it is limited by
            the range [0,1]. Defaults to 0.9. Ref: https://csaps.readthedocs.io/en/latest/formulation.html#definition
        min_samples: Minimum samples.
            Minimum number of data points required to form a distinct cluster. Defaults to 4.
            Defines the minimum number of data points required to form a distinct cluster. Rules of thumb for selecting
            the minimum samples value:

            * The larger the data set, the larger the value of MinPts should be.
            * If the data set is noisier, choose a larger value of MinPts Generally, MinPts should be greater than or
              equal to the dimensionality of the data set. For 2-dimensional data, use DBSCAN’s default value of 4
              (Ester et al., 1996).
            * If your data has more than 2 dimensions, choose MinPts = 2*dim, where dim= the dimensions of your data
              set (Sander et al., 1998).
        eps: Distance threshold.
            Defines the maximum distance between two samples for one to be considered as in the
            neighborhood of the other (i.e. belonging to the same cluster). This is the most important DBSCAN parameter
            to choose appropriately for your dataset and distance function. If no value is given, it is set automatically
            using Nearest Neighbors algorithm to calculate the average distance between each point and its k
            nearest neighbors, where k = min_samples (minimum samples). In ascending order on a k-distance graph, the
            optimal value for the threshold is at the point of maximum curvature (i.e. after plotting the average
            k-distances in where the graph has the greatest slope). This is not a maximum bound on the distances of
            points within a cluster.
            Defaults to None, eps value has to be > 0.0.
        time_window: Time window.
            Length of the time period to compute the rolling mean. The rolling mean and the data point value are the two features considered when calculating the distance to the furthest neighbour.
            This distance allows us to find the right epsilon when training dbscan. Defaults to '60min'.
            Accepted format: '3w', '10d', '5h', '30min', '10s'.
            If a number without unit (such as '60')is given, it will be considered as the number of minutes.
        del_zero_val: Remove zeros.
            Removes data points containing a value of 0. Defaults to False.

    Returns:
        pandas.Series: Time series without outliers.
    """
    outlier_indices = _get_outlier_indices(data, min_samples, eps, time_window, del_zero_val, reg_smooth)

    # Drop rows with outlier indices
    ts_without_indices = data.drop(outlier_indices)

    return ts_without_indices


@check_types
def _get_outlier_indices(
    data: pd.Series,
    min_samples,
    eps: Optional[float],
    time_window: pd.Timedelta,
    del_zero_val: bool,
    reg_smooth: float,
) -> pd.DatetimeIndex:
    """Get outlier indices.

    Identifies the indices of outliers in the given time series combining two methods, dbscan and csap.

    Args:
        data: Time series.
        reg_smooth: Smoothing factor.
        min_samples: Minimum samples.
        eps: Distance threshold.
        time_window: Time window.
            Accepted format: '3w', '10d', '5h', '30min', '10s'.
            If a number without unit (such as '60')is given, it will be considered as the number of minutes.
        del_zero_val: Remove zeros.
            Defaults to False.

    Returns:
        pandas.DatetimeIndex: An array of indices of outliers in the given time series.
    """
    # Handle special cases
    if len(data) <= 4:
        return pd.DatetimeIndex([])
    if (data.values == data.values[0]).all():  # if data is uniform, return 0 outliers
        return pd.DatetimeIndex([])

    # Validations
    validate_series_has_time_index(data)

    if eps is not None and eps <= 0.0:
        raise UserValueError("eps should be > 0.0.")

    df = data.to_frame()
    df = df.rename(columns={df.columns[0]: "val"})

    df["rolling_mean"] = df["val"].rolling(time_window, min_periods=1).mean()  # calculate features

    if del_zero_val:  # delete 0 values if user requests it
        df = df[df["val"] != 0.0]

    df_dbscan = df[["val"]].copy()
    validate_series_is_not_empty(df_dbscan)

    if eps is None:
        # calculate distance to the further neighbor in order to find best epsilon (radius) parameter for dbscan
        n_neighbors = min(len(df) - 3, 6)
        nbrs = NearestNeighbors(n_neighbors=n_neighbors).fit(df)
        distances, _ = nbrs.kneighbors(df)
        df["knn_dist"] = distances[:, -1]

        dist = np.sort(distances[:, -1])
        i = np.arange(len(dist))
        knee = KneeLocator(i, dist, S=1, curve="convex", direction="increasing", interp_method="polynomial")
        eps = dist[knee.knee] if dist[knee.knee] > 0 else 0.5  # if calculated eps is 0, use 0.5 as default value

    # train dbscan
    dbscan_model = DBSCAN(eps=eps, min_samples=min_samples).fit(df_dbscan)
    labels = dbscan_model.labels_

    outlier_positions = np.where(labels == -1)[0]

    # Get indices of outlier data points calculated using dbscan
    outlier_indices_dbscan = df.iloc[outlier_positions].index

    # Delete outliers detected by dbscan
    df_without_dbscan_outliers = df.drop(outlier_indices_dbscan)

    # Apply regression on the remaining data points
    df_reg = df.loc[df_without_dbscan_outliers.index, :]
    date_int = df_reg.index.to_series().astype(np.int64)
    date_int_stand = (date_int - date_int.mean()) / date_int.std()
    csaps_data = csaps(date_int_stand, df_reg["val"], date_int_stand, smooth=reg_smooth)

    # Delete points with high residuals
    res = pd.DataFrame(abs(df_reg["val"] - csaps_data))
    res["val_stand"] = (res["val"] - res["val"].mean()) / res["val"].std()
    res_stand_outliers = res[res["val_stand"] >= 3]
    # res_stand_no_outliers = res[res["val_stand"] < 3] reg_outliers_series = pd.Series(data =
    # res_stand_outliers.index, index= df.loc[res_stand_outliers.index,:]['val'].values) # todo return as output
    # df_without_outliers = df.loc[res_stand_no_outliers.index, :]

    outlier_indices_res_std = res_stand_outliers.index

    all_outliers = outlier_indices_dbscan.append(outlier_indices_res_std)

    return all_outliers


def outlier_percent(
    data: pd.Series,
    min_samples: int = 4,
    eps: Optional[float] = None,
    time_window: pd.Timedelta = pd.Timedelta("60min"),
    del_zero_val: bool = False,
    reg_smooth: float = 0.9,
) -> Optional[float]:
    """Calculates the percentage of outliers in the given time series.

    Args:
        data: Time series.
        min_samples: Minimum samples.
        eps: Distance threshold.
        time_window: Time window.
            Defaults to '60min'.
            Accepted format: '3w', '10d', '5h', '30min', '10s'.
            If a number without unit (such as '60')is given, it will be considered as the number of minutes.
        del_zero_val: Remove zeros.
            Defaults to False.
        reg_smooth: Smoothing factor.

    Returns:
        float: Percentage of outliers.
    """
    try:
        outlier_indices = _get_outlier_indices(data, min_samples, eps, time_window, del_zero_val, reg_smooth)
        return len(outlier_indices) / len(data) * 100
    except ZeroDivisionError:
        return None


OUTLIERS_REMOVE = remove_outliers



# Copyright 2022 Cognite AS
import numpy as np
import pandas as pd

from indsl.type_check import check_types

from ..exceptions import UserValueError


@check_types
def calculate_shutin_interval(
    shut_valve: pd.Series,
    min_shutin_len: int = 6,
    min_time_btw_shutins: int = 1,
    shutin_state_below_threshold: bool = True,
    shutin_threshold: int = None,
) -> pd.Series:
    """Shut-in interval.

    The shut-in interval is defined as the period when the valve is in closed state. The close state is determined based
    on the calculated manually-given threshold. The threshold is calculated based on the analysis of the valve signal
    histogram.

    Args:
        shut_valve: Shut-in valve.
            Time series with the shut-in valve opening.
        min_shutin_len: Minimum time.
            Minimum shut-in length in hours to be considered for detection.
        min_time_btw_shutins: Time between.
            Minimum time between consecutive shut-ins in hours to validate change of state.
        shutin_state_below_threshold: Below threshold.
            Indicator to tell the algorithm if the shut-in state is below the threshold.
        shutin_threshold: Threshold.
            Threshold between the valve open and close states. Defaults to None, meaning that the threshold is calculated.

    Returns:
        pandas.Series: Shut-in periods.
        Binary time series indicating open state or closed state: Open state= 1, Close state = 0.

    Raises:
        RuntimeError: If threshold cannot be determined in the evaluation window, the user is asked to increase the evaluation period. Recommended period is 30 days.
    """
    # calculate threshold
    if not shutin_threshold:
        shutin_threshold = calculate_threshold(shut_valve)

    # create binary timeseries
    wvalve = shut_valve.to_frame(name="valve_state_org")
    wvalve["timestamp"] = wvalve.index
    if shutin_state_below_threshold:
        wvalve["valve_state"] = np.where(wvalve["valve_state_org"] > shutin_threshold, 1, 0)
    else:
        wvalve["valve_state"] = np.where(wvalve["valve_state_org"] > shutin_threshold, 0, 1)
    wvalve["valve_diff"] = wvalve["valve_state"].diff()
    wvalve["final_indicator"] = 1
    wvalve = wvalve.dropna()

    # get shut-in starts and ends
    sh_start = list(wvalve[wvalve["valve_diff"] == -1]["timestamp"])
    sh_end = list(wvalve[wvalve["valve_diff"] == 1]["timestamp"])
    # create time pairs
    if sh_start and sh_end:
        # case 1: several shut-ins exist - the shut-in comes from the past - make a pseudo-start of the shut-in
        if sh_end[0] < sh_start[0]:
            sh_start = [wvalve.iloc[0, :]["timestamp"]] + sh_start
        # case 2: several shut-ins exist - the shut-in goes to the future - make a pseudo-end of shut-in
        if sh_end[-1] < sh_start[-1]:
            sh_end = sh_end + [wvalve.iloc[-1, :]["timestamp"]]
    elif not sh_start:
        # case 3: only one shut-in exist - the shut-in comes from the past - make a pseudo-start of the shut-in
        sh_start = [wvalve.iloc[0, :]["timestamp"]] + sh_start
    else:
        # case 4: only one shut-in exist - the shut-in comes from the past - make a pseudo-start of the shut-in
        sh_end = sh_end + [wvalve.iloc[-1, :]["timestamp"]]
    time_pairs = list(zip(sh_start, sh_end))

    # merge the shut-ins based on the minimum distance between consecutive shut-ins
    time_pairs_upd = []
    time_pairs.append(time_pairs[-1])
    i = 0
    new_pair = time_pairs[i]
    time_pairs_upd.append(new_pair)
    while i < len(time_pairs) - 1:
        delta_hr = (time_pairs[i + 1][0] - new_pair[1]).total_seconds() / 3600
        if delta_hr < min_time_btw_shutins:
            new_pair = (new_pair[0], time_pairs[i + 1][1])
            if time_pairs_upd[-1][0] == new_pair[0]:
                time_pairs_upd[-1] = new_pair
        else:
            new_pair = time_pairs[i + 1]
            time_pairs_upd.append(new_pair)
        i = i + 1

    # filter the shut-ins based on the minimum duration condition
    time_pairs_upd = [pair for pair in time_pairs_upd if (pair[1] - pair[0]).total_seconds() / 3600 > min_shutin_len]
    for start, end in time_pairs_upd:
        mask = np.logical_and(
            wvalve["timestamp"] >= start,
            wvalve["timestamp"] <= end,
        )
        wvalve.loc[mask, "final_indicator"] = 0
    return wvalve["final_indicator"].dropna().rename(None)


@check_types
def calculate_threshold(shut_valve: pd.Series) -> float:
    """Calculate valve threshold.

    The valve threshold refers to the threshold between valve open and close states. The threshold is calculated as
    a mean of two histogram peaks (max values). The histogram should have two peaks indicating two states.

    Args:
        shut_valve: pandas.Series
            Shut-in valve signal

    Returns:
        float: Threshold

    Raises:
        RuntimeError: If threshold cannot be determined in the evaluation window, the user is asked to increase the evaluation period. Recommended period is 30 days.
    """
    freqs, edges = np.histogram(shut_valve.values, bins=5)
    # get indices of 2 most frequent bins
    freqs_sorted = [el[0] for el in sorted(enumerate(freqs), key=lambda x: x[1])]
    max1_idx, max2_idx = freqs_sorted[-2], freqs_sorted[-1]
    # get mean values between bin edges
    mean_edge = np.array([(edges[i] + edges[i + 1]) / 2 for i in range(len(edges) - 1)])
    # calculate threshold
    if 1.5 * min(mean_edge[max1_idx], mean_edge[max2_idx]) > max(mean_edge[max1_idx], mean_edge[max2_idx]):
        raise UserValueError("Not enough data to detect the threshold")
    else:
        threshold = (mean_edge[max1_idx] + mean_edge[max2_idx]) / 2
    return threshold



# Copyright 2021 Cognite AS
import numpy as np
import pandas as pd

from indsl.exceptions import UserRuntimeError
from indsl.resample.auto_align import auto_align
from indsl.type_check import check_types


@check_types
def status_flag_filter(
    data: pd.Series, filter_by: pd.Series, int_to_keep: int = 0, align_timesteps: bool = False
) -> pd.Series:
    """Status flag filter.

    This function filters any given data series by a series with integers denoting different states. A typical example of
    such a series is a series of 0 and 1 where 1 would indicate the presence of an anomaly.
    The status flag filter retrieves all relevant indices and matches these to the data series.

    Args:
        data: Time series.
        filter_by: Status flag time series.
            Time series values are expected to be integers. If not, values are cast to integer automatically.
        int_to_keep: Value.
            Value to filter by in the boolean filter
        align_timesteps: Auto-align.
          Automatically align time stamp of input time series. Default is False.

    Returns:
        pandas.Series: Filtered time series

    Raises:
        UserRuntimeError: Time series returns no data. This could be due to insufficient data in either `data`
            or `filter_by`, or `filter_by` series contains no values of `int_to_keep`.
    """
    data, filter_by = auto_align([data, filter_by], align_timesteps)

    # Select index to keep
    filter_index = filter_by[filter_by == int_to_keep].index

    # Match data index to filter index
    data_index = data.index[data.index.isin(filter_index)]

    # Check for no entries
    if len(data_index) == 0:
        raise UserRuntimeError("Current filter returns no entries for data.")

    # Check the filter_by contains integers
    if filter_by.dtypes is not np.dtype("int64"):
        filter_by = filter_by.astype(np.int64)
        # warnings.warn(
        #    "Input data `filter_by` contains types other than integers. Casting values to int and continuing!",
        #    RuntimeWarning,
        # )

    return data[data_index]



# Copyright 2021 Cognite AS
import numpy as np

from indsl.resample.auto_align import auto_align
from indsl.type_check import check_types


def exp(x):
    """Exp.

    Calculates the exponential of a time series.

    Args:
        x: time-series

    Returns:
        pandas.Series: time series
    """
    return np.exp(x)


def log(x):
    """Ln.

    Calculates the natural logarithm of a time series.

    Args:
        x: time-series

    Returns:
        pandas.Series: time series
    """
    return np.log(x)


def log2(x):
    """Log base 2.

    Calculates the logarithm with base 2 of a time series.

    Args:
        x: time-series

    Returns:
        pandas.Series: time series
    """
    return np.log2(x)


def log10(x):
    """Log base 10.

    Calculates the logarithm with base 10 of a time series.

    Args:
        x: time-series

    Returns:
        pandas.Series: time series
    """
    return np.log10(x)


@check_types
def logn(x, base, align_timesteps: bool = False):
    """Log, any base.

    Calculates the logarithm with base “n” of a time series.

    Args:
        x: Input time-series or number
        base: Base time-series or number
        align_timesteps (bool) : Auto-align
          Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: time series
    """
    x, base = auto_align([x, base], align_timesteps)
    return np.log(x) / np.log(base)



# Copyright 2022 Cognite AS
from typing import Optional

import numpy as np
import pandas as pd

from numba import jit

from indsl import versioning
from indsl.ts_utils.ts_utils import time_parse
from indsl.type_check import check_types


@versioning.register(version="1.0", deprecated=True)
@check_types
def lwma(
    data: pd.Series,
    time_window: str = "60min",
    min_periods: int = 1,
    resample: bool = False,
    resample_window: str = "60min",
) -> pd.Series:
    """Linear weighted moving average.

    The linear weighted moving average gives more weight to the more recent observations and gradually less to the older
    ones.

    Args:
        data: Time series.
        time_window: Time window.
            Length of the time period to compute the rolling mean. Defaults to '60min'.
            If the user gives a number without unit (such as '60'), it will be considered as the number of minutes.
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.
        min_periods: Minimum samples.
            Minimum number of observations in the time window required to estimate a value (otherwise result is NA).
            Defaults to 1.
        resample: Resample.
            Resamples the calculated linear weighted moving average series. Defaults to False
        resample_window: Resampling window.
            Time window used to resample the calculated linear weighted moving average series. Defaults to '60min'.

    Returns:
        pandas.Series: Smoothed time series.
    """
    time_window = time_parse(time_window)
    lwma = np.zeros_like(data)
    win_pts_lst = np.array(data.rolling(time_window).count().values.astype(int))

    for i in range(len(data)):
        lwma[i] = get_lwma_val(data.values, win_pts_lst, i, min_periods=min_periods)
    lwma_series = pd.Series(lwma, index=data.index)
    if resample:
        resample_window = time_parse(resample_window)
        lwma_series = lwma_series.resample(resample_window).mean()
    return lwma_series


@jit(nopython=True)
def get_lwma_val(data: np.ndarray, win_pts_lst: np.ndarray, i: int, min_periods: int) -> Optional[float]:
    """Calculates the Linear Weighted Moving Average for the current datapoint.

    Args:
        data (numpy.ndarray): Data values of the timeseries.
        win_pts_lst (numpy.ndarray): Array with the number of points for each time window.
        i (int): Index of the current value.
        min_periods (int): Minimum number of observations in window required to have a value (otherwise result is NA).

    Returns:
        None: Length of data was smaller than minimum number of observations in window required to have a value.
        float: Calculated lwma for the current datapoint.
    """
    if len(data) < min_periods:
        return None
    else:
        win_pts = win_pts_lst[i]
        win_values = data[(i + 1) - win_pts : i + 1]
        weights = np.linspace(1, min(len(data), win_pts), min(len(data), win_pts))
        lwma_val = np.dot(win_values, weights) / weights.sum()
    return lwma_val



# Copyright 2022 Cognite AS
from typing import Union

import numpy as np
import pandas as pd

from indsl.resample.auto_align import auto_align
from indsl.ts_utils.ts_utils import scalar_to_pandas_series
from indsl.type_check import check_types


@check_types
def total_head(
    discharge_pressure: Union[pd.Series, float],
    suction_pressure: Union[pd.Series, float],
    den: Union[pd.Series, float],
    align_timesteps: bool = False,
) -> pd.Series:
    r"""Total head.

    Head is a measure of the potential of a liquid to reach a certain
    height. The head is essentially a unit of pressure. The total head
    is the difference in pressure of the discharge to the suction of
    the pump.he formula for total head :math:`h` [m] given inputs discharge pressure
    :math:`P_{discharge}` [Pa], suction pressure :math:`P_{suction}` [Pa] and liquid density
    :math:`\rho_L\:[\frac{kg}{m^3}]`.

    .. math::
        h = \frac{P_{discharge} - P_{suction}}{9.81\rho_L}

    Args:
        discharge_pressure: Discharge pressure [Pa].
            Discharge pressure of a centrifugal pump.
        suction_pressure: Suction pressure [Pa].
            Suction pressure of a centrifugal pump.
        den: Density of the fluid [:math:`\mathrm{\frac{kg}{m^3}}`].
        align_timesteps: Auto-align.
            Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Total head [m]
            Difference in total discharge head and the total suction head.

    """
    # auto-align
    discharge_pressure, suction_pressure, den = auto_align([discharge_pressure, suction_pressure, den], align_timesteps)
    head = (discharge_pressure - suction_pressure) / (den * 9.81)
    return scalar_to_pandas_series(head)


@check_types
def percent_BEP_flowrate(
    pump_liquid_flowrate: Union[pd.Series, float], BEP_flowrate: Union[pd.Series, float], align_timesteps: bool = False
) -> pd.Series:
    r"""BEP from flowrate [%].

    Centrifugal pumps operate optimally at a specific liquid flow rate which is typically called the Best Efficiency Point (BEP).
    This function calculates the flow rate relative to BEP as a percentage.
    i.e. 100% means the current flow rate is at the BEP, 110% means the
    current flow rate is 10% above BEP. The formula for this equation is:

    .. math::
        BEP\:from\:flow rate\:[\%]=\frac{Pump\:liquid\:flow rate}
        {BEP}*100

    Args:
        pump_liquid_flowrate: Pump liquid flow rate [:math:`\mathrm{\frac{m^3}{s}}`].
            The current flow rate of the pump.
        BEP_flowrate: Best efficiency point [-].
            The best efficiency flow rate point of the pump.
        align_timesteps: Auto-align.
            Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: BEP to current flowrate [%].
            Percentage of current flow rate to BEP

    """
    pump_liquid_flowrate, BEP_flowrate = auto_align([pump_liquid_flowrate, BEP_flowrate], align_timesteps)
    percent_BEP_Q = pump_liquid_flowrate / BEP_flowrate * 100

    return scalar_to_pandas_series(percent_BEP_Q)


@check_types
def pump_hydraulic_power(
    pump_liquid_flowrate: Union[pd.Series, float],
    total_head: Union[pd.Series, float],
    den: Union[pd.Series, float],
    align_timesteps: bool = False,
) -> pd.Series:
    r"""Pump hydraulic power.

    Pump hydraulic power [W] is the amount of energy per unit time
    delivered to the liquid. Pump hydraulic power can be calculated
    if the pump liquid flow rate :math:`Q_L\:[\frac{m^3}{s}]`, total head across the pump
    :math:`h` [m], and density of the fluid :math:`\rho_L\:[\frac{kg}{m^3}]`.

    .. math::
        Pump\:hydraulic\:power=9.81Q_L\rho_Lh

    Args:

        pump_liquid_flowrate: Pump liquid flow rate [:math:`\mathrm{\frac{m^3}{s}}`].
            The current flow rate of the pump.
        total_head: Total head across pump [m].
            Difference in pressure between discharge and suction of pump.
        den: Density of the fluid [:math:`\mathrm{\frac{kg}{m^3}}`].
        align_timesteps: Auto-align.
            Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Pump hydraulic power [W].
            Pump hydraulic power of pump.

    """
    # auto-align
    pump_liquid_flowrate, total_head, den = auto_align([pump_liquid_flowrate, total_head, den], align_timesteps)
    P = pump_liquid_flowrate * den * 9.81 * total_head

    return scalar_to_pandas_series(P)


@check_types
def pump_shaft_power(
    pump_hydraulic_power: Union[pd.Series, float],
    pump_liquid_flowrate: Union[pd.Series, float],
    eff_parameter_1: Union[pd.Series, float],
    eff_parameter_2: Union[pd.Series, float],
    eff_intercept: Union[pd.Series, float],
    align_timesteps: bool = False,
) -> pd.Series:
    r"""Pump shaft power.

    Pump shaft power is the input power delivered by the shaft.
    Pump shaft power can be calculated by dividing the pump hydraulic hp
    by the pump efficiency. Pump efficiency is a function of liquid flow rate :math:`\eta(Q_L)`.
    The pump efficiency curve as a function of liquid flow rate is assumed to be a 2nd order polynomial.
    Therefore the input parameters of the curve are coefficients
    to :math:`x^2` and :math:`x` and the :math:`y` intercept of the curve.

    .. math::
        Pump\:shaft\:power=\frac{Pump\:hydraulic\:power}{\eta(Q_L)}

    Args:

        pump_hydraulic_power: Pump hydraulic power [W].
        pump_liquid_flowrate: Pump liquid flowrate [:math:`\mathrm{\frac{m^3}{h}}`].
            The current flow rate of the pump.
        eff_parameter_1: :math:`x^2` coefficient [-].
            Coefficient of :math:`x^2`.
        eff_parameter_2: :math:`x` coefficient [-].
            Coefficient of :math:`x`.
        eff_intercept: :math:`y`-intercept [-].
            Coefficient of :math:`y`-intercept of curve
        align_timesteps: Auto-align.
            Automatically align time stamp  of input time series. Default is False.

    Returns:
        pandas.Series: Pump shaft power [W]
            Pump shaft power of pump.
    """
    # auto-align
    pump_liquid_flowrate, pump_hydraulic_power, eff_parameter_1, eff_parameter_2, eff_intercept = auto_align(
        [pump_liquid_flowrate, pump_hydraulic_power, eff_parameter_1, eff_parameter_2, eff_intercept], align_timesteps
    )

    p = (eff_parameter_1, eff_parameter_2, eff_intercept)
    eff = np.polyval(p, pump_liquid_flowrate) / 100

    P = pump_hydraulic_power / eff

    return scalar_to_pandas_series(P)


@check_types
def recycle_valve_power_loss(
    Q_valve: Union[pd.Series, float],
    total_head: Union[pd.Series, float],
    den: Union[pd.Series, float],
    align_timestamps: bool = False,
) -> pd.Series:
    r"""Pump recycle valve power loss.

    This calculation can be used where there is a recirculation line with a recycle valve whose purpose is to maintain a minimum flow through the pump.
    The calculation does not consider the difference in pump efficiency at different flow rates. This is acceptable because pumps are usually sized to account for extra flow due to recirculation.

    .. math::
            Power\:loss=9.81Q_{valve}\rho_Lh

    Args:
        Q_valve: Valve flow rate [:math:`\mathrm{\frac{m^3}{h}}`].
            Flow rate through the recycle valve.
        total_head: Pump total head [m].
        den: Density of the fluid [:math:`\mathrm{\frac{kg}{m^3}}`].
        align_timestamps: Auto-align.
            Automatically align time stamp  of input time series. Defaults to False.

    Returns:
        pd.Series: Power loss [W].
            Power loss by recirculation though the pump.
    """
    return pump_hydraulic_power(Q_valve, total_head, den, align_timestamps)



from typing import List

import pandas as pd

from ..type_check import check_types


@check_types
def univariate_polynomial(signal: pd.Series, coefficients: List[float] = [0.0, 1.0]) -> pd.Series:
    """Univariate polynomial.

    Creates a univariate polynomial :math:`y`, of degree :math:`n`, from the time series :math:`x`, and a list of
    coefficients :math:`a_{n}`:

    .. math::

        y = a_0 + a_1x + a_2x^2 + a_3x^3 + ... + a_nx^n

    Args:
        signal: Time series
        coefficients: Coefficients.
            List of coefficients separated by semicolons. The numbers must be entered, separated by semicolons (e.g., 0; 1).
            The default is :math:`0.0; 1.0`, which returns the original time series.

    Returns:
        pandas.Series: Output
    """
    poly = pd.Series(0, index=signal.index)
    n = 0
    for a in coefficients:
        poly = poly + a * signal**n
        n = n + 1

    return poly



class UserException(Exception):
    """Exception raised when the error is due to the user."""

    pass


class UserValueError(ValueError, UserException):
    """Exception raised when the user provides a wrong value."""

    pass


class UserTypeError(TypeError, UserException):
    """Exception raised when the user provides a value with the wrong type."""

    pass


class UserRuntimeError(RuntimeError, UserException):
    """Exception raised when the error in runtime is due to the user."""

    pass



# Copyright 2021 Cognite AS
from typing import Literal

import numpy as np
import pandas as pd

from indsl import versioning
from indsl.type_check import check_types

from . import drift_detector_v1  # noqa


@versioning.register(version="2.0", changelog="Updated input parameter types")
@check_types
def drift(
    data: pd.Series,
    long_interval: pd.Timedelta = pd.Timedelta("3d"),
    short_interval: pd.Timedelta = pd.Timedelta("4h"),
    std_threshold: float = 3,
    detect: Literal["decrease", "increase", "both"] = "both",
) -> pd.Series:
    """Drift.

    This function detects data drift (deviation) by comparing two rolling averages, short and long interval, of the signal. The
    deviation between the short and long term average is considered significant if it is above a given threshold
    multiplied by the rolling standard deviation of the long term average.

    Args:
        data: Time series.
        long_interval: Long length.
            Length of long term time interval.
        short_interval: Short length.
            Length of short term time interval.
        std_threshold: Threshold.
            Parameter that determines if the signal has changed significantly enough to be considered drift. The threshold
            is multiplied by the long term rolling standard deviation to take into account the recent condition of the
            signal.
        detect: Type.
            Parameter to determine if the model should detect significant decreases, increases or both. Options are:
            "decrease", "increase", or "both". Defaults to "both".

    Returns:
      pandas.Series: Boolean time series.
        Drift = 1, No drift = 0.
    """
    # Compute long term average and std
    long_average = data.rolling(long_interval).mean()
    long_stds = data.rolling(long_interval).std()

    # Compute short term average
    short_average = data.rolling(short_interval).mean()

    # Calculate drift masks
    drift_mask_increase = short_average > long_average + std_threshold * long_stds
    drift_mask_decrease = short_average < long_average - std_threshold * long_stds

    # set mask according to setting
    if detect == "increase":
        drift_mask = drift_mask_increase
    elif detect == "decrease":
        drift_mask = drift_mask_decrease
    else:
        drift_mask = np.logical_or(drift_mask_increase, drift_mask_decrease)

    return drift_mask.dropna().astype(int)



# Copyright 2022 Cognite AS
from typing import Literal

import pandas as pd

from skimage.restoration import denoise_wavelet

from indsl import versioning

from ..exceptions import UserValueError
from ..type_check import check_types
from . import wavelet_filter_v1  # noqa


# TODO: Add more wavelet types from https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html
wavelet_options = Literal[
    "db1",
    "db2",
    "db3",
    "db4",
    "db5",
    "db6",
    "db7",
    "db8",
    "sym2",
    "sym3",
    "sym4",
    "coif1",
    "coif2",
    "coif3",
    "coif4",
    "coif5",
]


@versioning.register(version="2.0", changelog="updated wavelet option type, added coif[1-5] as wavelet option")
@check_types
def wavelet_filter(data: pd.Series, level: int = 2, wavelet: wavelet_options = "db8") -> pd.Series:
    """Wavelet de-noising.

    Filtering industrial data using wavelets can be very powerful as it uses a *dual* frequency-time
    representation of the original signal, which allows separating noise frequencies from valuable signal frequencies.
    For more on wavelet filter or other application, see https://en.wikipedia.org/wiki/Wavelet

    Args:
        data: Time series.
            The data to be filtered. The series must have a pandas.DatetimeIndex.
        level: Level.
            The number of wavelet decomposition levels (typically 1 through 6) to use.
        wavelet: Type.
            The default is a Daubechies wavelet of order 8 (*db8*). For other types of wavelets, see the
            `pywavelets package <https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html>`_.
            The thresholding methods assume an orthogonal wavelet transform and may not choose the threshold
            appropriately for biorthogonal wavelets. Orthogonal wavelets are desirable because white noise in
            the input remains white noise in the sub-bands. Therefore one should choose one of the db[1-20], sym[2-20],
            or coif[1-5] type wavelet filters.

    Raises:
        UserValueError: The level value needs to be a positive integer
        UserValueError: The level value can not exceed the length of data points

    Returns:
        pandas.Series: Filtered time series.
    """
    # TODO: Add more info about wavelet type input in docstrings
    if level <= 0:
        raise UserValueError("The level value needs to be a positive integer")
    if level >= len(data.values):
        raise UserValueError("The level value can not exceed the length of data points")
    res = denoise_wavelet(
        data, wavelet_levels=level, wavelet=wavelet, method="VisuShrink", mode="soft", rescale_sigma=True
    )
    return pd.Series(res, index=data.index)



# Copyright 2022 Cognite AS
from .arma_predictor import arma_predictor
from .holt_winters_predictor import holt_winters_predictor


TOOLBOX_NAME = "Forecast"

__all__ = ["arma_predictor", "holt_winters_predictor"]



# Copyright 2021 Cognite AS
from typing import Optional

import numpy as np
import pandas as pd

from numba import jit

from indsl import versioning

from ..type_check import check_types
from ..validations import validate_series_is_not_empty
from . import eweight_ma_v1  # noqa


@versioning.register(version="2.0", changelog="Update parameter type")
@check_types
def ewma(
    data: pd.Series,
    time_window: pd.Timedelta = pd.Timedelta("60min"),
    min_periods: int = 1,
    adjust: bool = True,
    max_pt: int = 200,
    resample: bool = False,
    resample_window: pd.Timedelta = pd.Timedelta("60min"),
) -> pd.Series:
    """Exp. weighted moving average.

    The exponential moving average gives more weight to the more recent observations. The weights fall exponentially
    as the data point gets older. It reacts more than the simple moving average with regards to recent movements.
    The moving average value is calculated following the definition yt=(1−α)yt−1+αxt if adjust = False or
    yt=(xt+(1−α)*xt−1+(1−α)^2*xt−2+...+(1−α)^t*x0) / (1+(1−α)+(1−α)^2+...+(1−α)^t) if adjust = True.

    Args:
        data: Time series.
            Data with a pd.DatetimeIndex.
        time_window: Time window.
            Defines how important the current observation is in the calculation of the EWMA. The longer the period, the slowly it adjusts to reflect current trends. Defaults to '60min'.
            Accepted string format: '3w', '10d', '5h', '30min', '10s'.
            The time window is converted to the number of points for each of the windows. Each time window may have different number of points if the timeseries is not regular.
            The number of points specify the decay of the exponential weights in terms of span α=2/(span+1), for span≥1.
        min_periods: Minimum number of data points.
            Minimum number of data points inside a time window required to have a value (otherwise result is NA). Defaults to 1.
            If min_periods > 1 and adjust is False, the SMA is computed for the first observation.
        adjust: Adjust.
            If true, the exponential function is calculated using weights w_i=(1−α)^i.
            If false, the exponential function is calculated recursively using yt=(1−α)yt−1+αxt. Defaults to True.
        max_pt: Maximum number of data points.
            Sets the maximum number of points to consider in a window if adjust = True. A high number of points will require more time to execute. Defaults to 200.
        resample: Resample.
            If true, resamples the calculated exponential moving average series. Defaults to False.
        resample_window: Resampling window
            Time window used to resample the calculated exponential moving average series. Defaults to '60min'.

    Returns:
        pandas.Series: Smoothed time series.
    """
    validate_series_is_not_empty(data)

    win_pts_lst = np.array(data.rolling(time_window).count().values)
    ewma_vals = np.zeros_like(data)
    a = 2.0 / (win_pts_lst + 1)
    if not adjust:
        for i, value in enumerate(data):
            ewma_vals[i] = ewma_pt_not_adjust(data.values, value, min_periods, ewma_vals, i, a)
    else:
        ewma_vals[0] = data[0]
        for i, value in enumerate(data):
            if i == 0:
                continue
            ewma_vals[i] = ewma_pt_adjust(data.values, value, i, a, max_pt)
    bool_mask = win_pts_lst < min_periods
    ewma_vals[bool_mask] = None
    ewma_series = pd.Series(ewma_vals, index=data.index)
    if resample:
        ewma_series = ewma_series.resample(resample_window).mean()
    return ewma_series


@jit(nopython=True)  # @njit
def ewma_pt_adjust(data: np.ndarray, value: float, i: int, a: np.ndarray, max_pt: int = 200) -> float:
    """Calculates ewma value using weights wi=(1−a)^i.

    Args:
        data (numpy.ndarray): Data values of the time series.
        value (float): Current datapoint.
        i (int): Index of the current value.
        a (np.ndarray): Alpha value to compute ewma.
        max_pt (int): Maximum number of points to consider when calculating ewma. Defaults to 200.

    Returns:
        float: Calculated ewma for the current datapoint.
    """
    values = data[:i] if len(data[:i]) <= max_pt else data[i - max_pt : i]
    exp = np.arange(1, len(values) + 1)[::-1]
    w_den = (1 - a[i]) ** exp
    w_nom = w_den * values
    ewma_val = (value + w_nom.sum()) / (1 + w_den.sum())
    return ewma_val


# @jit(nopython=True)
def ewma_pt_not_adjust(
    data: np.ndarray, value: float, min_periods: int, ewma_vals: np.ndarray, i: int, a: np.ndarray
) -> Optional[float]:
    """Calculates the ewma recursively using yt=(1−α)yt−1+αxt.

    Args:
        data (numpy.ndarray): Data values of the time series.
        value (float): Current datapoint.
        min_periods (int): Minimum number of observations in window required to have a value (otherwise result is NA).
            If min_periods > 1 the SMA is computed for the first observation.
        ewma_vals (numpy.ndarray): Array of the ewma values.
        i (int): Index of the current value.
        a (np.ndarray): Alpha value to compute ewma.

    Returns:
        None: Index of the current value is smaller than minimum number of observations in window required to have a value.
        float: Calculated ewma for the current datapoint.
    """
    if i < min_periods - 1:
        return None
    elif i == min_periods - 1:  # first value is the simple moving average
        start_sma = data[:min_periods].sum() / len((data[:min_periods]))
        return start_sma
    else:
        ewma_val = (a[i] * value) + ((1 - a[i]) * ewma_vals[i - 1])
        return ewma_val



# Copyright 2022 Cognite AS
from typing import List, Union

import numpy as np
import pandas as pd

from ..exceptions import UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_time_index, validate_series_is_not_empty


@check_types
def alma(data: pd.Series, window: int = 10, sigma: float = 6, offset_factor: float = 0.75) -> pd.Series:
    """Arnaud Legoux moving average.

    Moving average typically used in the financial industry, which aims to strike a good balance between smoothness
    and responsiveness (i.e., capture a general smoothed trend without allowing for significant lag). It can be
    interpreted as a Gaussian weighted moving average with an offset, where the offset, spread, and window size are
    user-defined.

    Args:
        data: Time series.
        window: Window size.
            Defaults to 10 data points or time steps for uniformly sample time series.
        sigma: Sigma.
            Parameter that controls the width of the Gaussian filter. Defaults to 6.
        offset_factor: Offset factor.
            Parameter that controls the magnitude of the weights for each past observation within the window.
            Defaults to 0.75.

    Raises:
        UserValueError: Time series is empty.
        UserValueError: Data has to be a time series.
        UserValueError: Not enough data to perform the calculation.

    Returns:
        pandas.Series: Smoothed data.
    """
    # TODO : Refactor to accept TIME window instead of number of data points

    validate_series_is_not_empty(data)
    validate_series_has_time_index(data)

    # Check data
    if len(data) <= window:
        raise UserValueError(f"Not enough data to perform calculation. Expected {window} but got {len(data)}")

    # Check inputs
    if window == sigma == 0:
        raise UserValueError(
            "window or sigma can't be zero. Please change these user defined values to positive values."
        )

    # Calculate weights
    offset = int(offset_factor * window)
    k = np.array(range(0, window))
    weights = np.exp(-((k - offset) ** 2) / (sigma**2))

    # Apply smoothing function
    res = data.rolling(window=window).apply(lambda x: calculate_alma(x, weights))

    return res.dropna()


@check_types
def calculate_alma(values: Union[List, pd.Series], weights: np.ndarray) -> float:
    """Calculate alma value for a window time.

    Args:
        values: Datapoints in the window time.
        weights: Weights to calculate Alma value.

    Returns:
        float: Calculated Alma value.
    """
    weighted_sum = weights * values
    alma = weighted_sum.sum() / weights.sum()
    return alma



# Copyright 2022 Cognite AS
from numbers import Number
from typing import Any, Iterable, List, Union

import numpy as np
import pandas as pd

from indsl.type_check import check_types

from .reindex import reindex_many


def _is_timeseries(data: Any) -> bool:
    return isinstance(data, pd.Series) and np.issubdtype(data.index, np.datetime64)


@check_types
def auto_reindex(data: List[Union[Iterable, Number, pd.Series]], enabled: bool = True) -> list:
    """Automatically re-reindex input time series."""
    if not enabled:
        return data

    # Extract sub-list with time-series - only these entries will be reindexed
    id_ts_map = {i: d for i, d in enumerate(data) if _is_timeseries(d)}

    # Reindex
    reindexed_ts = reindex_many(list(id_ts_map.values()), bounded=True)

    out = data.copy()
    for idx, ts in zip(id_ts_map.keys(), reindexed_ts):
        out[idx] = ts
    return out


@check_types
def auto_align(data: List[Union[Iterable, Number, pd.Series]], enabled: bool = True) -> list:
    """Automatically re-reindex input time series."""
    return auto_reindex(data, enabled=enabled)



# Copyright 2022 Cognite AS
from .simple_filters import status_flag_filter
from .trend import trend_extraction_hilbert_transform
from .wavelet_filter import wavelet_filter


__all__ = ["wavelet_filter", "status_flag_filter", "trend_extraction_hilbert_transform"]

TOOLBOX_NAME = "Filter"



# Copyright 2022 Cognite AS
import numpy as np
import pandas as pd

from indsl.type_check import check_types


@check_types
def Haaland(Re: pd.Series, roughness: float) -> pd.Series:
    """Haaland equation.

    The `Haaland equation <https://en.wikipedia.org/wiki/Darcy_friction_factor_formulae#Haaland_equation>`_ was
    proposed in 1983 by Professor S.E. Haaland of the Norwegian Institute of Technology.
    It is used to directly solve the Darcy–Weisbach friction factor for a full-flowing circular pipe. It is an
    approximation of the implicit Colebrook–White equation, but the discrepancy from experimental data is well within
    the accuracy of the data.

    Args:
        Re: Reynolds Number
        roughness: Surface roughness

    Returns:
        pandas.Series: Friction factor
    """
    den = -1.8 * np.log10((6.9 / Re[Re != 0]) + (roughness / 3.7) ** 1.1)
    return 1 / den**2



# Copyright 2022 Cognite AS
from typing import List, Literal

import numpy as np
import pandas as pd
import pandas.core.indexes.datetimes
import scipy

from indsl import versioning
from indsl.ts_utils.ts_utils import functional_mean

from ..exceptions import UserTypeError, UserValueError
from ..type_check import check_types
from . import reindex_v1  # noqa


kind_options = Literal["pointwise", "average"]
method_options = Literal["zero", "next", "slinear", "quadratic", "cubic"]


@versioning.register(version="2.0", changelog="update data types")
@check_types
def reindex(
    data1: pd.Series,
    data2: pd.Series,
    method: method_options = "slinear",
    kind: kind_options = "pointwise",
    bounded: bool = False,
) -> pd.Series:
    """Reindex.

    This method offers data reindexing onto a common index and fills missing data points.
    If bounded is false, the common index is the union of the the input time-series indices.
    If bounded is true, the common index is restricted to the period where the time-series overlap.
    All not-a-number (NaN) values are removed in the output time series.

    Args:
        data1: First time series.

        data2: Second time series.

        method: Method.
            Specifies the interpolation method. Defaults to "linear". Possible inputs are :

            * 'zero': zero order spline interpolation with forward filling mode, i.e., the previous known value of any point is used.
            * 'next': zero order spline interpolation with backward filling mode, i.e., the next known value of any point is used.
            * 'slinear': linear order spline interpolation.
            * 'quadratic': quadratic order spline interpolation.
            * 'cubic': cubic order spline interpolation.

        kind: Kind.
            Specifies the kind of returned data points. Defaults to "pointwise".  Possible inputs are:

            * 'pointwise': returns the pointwise value of the interpolated function for each timestamp.
            * 'average': returns the average of the interpolated function within each time period.

        bounded: Bounded.
            Specifies if the output should be bounded to avoid extrapolation. Defaults to False. Possible inputs are:

            * True: Return the intersection of the time periods of the input time series.
            * False: Return the union of the time periods of the input time series. Extrapolate points outside of the data range.


    Returns:
        pandas.Series: First reindexed time series
            Reindexed time series with common indices.

    Raises:
        UserValueError: All time series must have at least two values
    """
    return reindex_many([data1, data2], method, kind, bounded)[0]


@versioning.register(version="2.0", changelog="update data types")
@check_types
def reindex_many(
    data: List[pd.Series],
    method: method_options = "slinear",
    kind: kind_options = "pointwise",
    bounded: bool = False,
) -> List[pd.Series]:
    """Re-indexes list.

    Args:
        data: List[pd.Series]
        method: Method, defaults to Method.LINEAR
        kind: Kind, defaults to Kind.POINTWISE
        bounded: bool, defaults to False

    Raises:
        UserTypeError: Only DateTimeIndex index supported.
        UserValueError: One of the time series has less than two values.
        UserTypeError:  The re-indexed time series is empty.

    Returns:
        List[pd.Series]: re-indexed list
    """
    # Nothing to do if data contains less than 2 series
    if len(data) < 2:
        return data

    # Return if all indices are already aligned and
    # none of the input series contain NaNs
    indices = [d.index for d in data]
    if all([indices[0].equals(idx) for idx in indices[1:]]) and not any([d.isnull().values.any() for d in data]):
        return data

    # check input types
    if not all(isinstance(d.index, pd.DatetimeIndex) for d in data):
        raise UserTypeError("Only DateTimeIndex index supported.")

    # Check validity of input time series
    if any([len(d.dropna()) < 2 for d in data]):
        raise UserValueError("One of the time series has less than two values.")

    # If bounded=True, we need to remove Nan values here.
    # Otherwise, we might end up extrapolating if there are NaN values at
    # the beginning/end of a time-series (which is in conflict with the bounded=True flag)
    if bounded:
        data = [d.dropna() for d in data]
        indices = [d.index for d in data]

    # Get the union of the indices
    idx_common = indices[0]
    for index in indices[1:]:
        idx_common = idx_common.union(index)

    # If bounded is true, restrict the dataset to a range where both time-series have data
    if bounded:
        start = max([idx.min() for idx in indices])
        stop = min([idx.max() for idx in indices])

        start_loc = idx_common.get_loc(start)
        stop_loc = idx_common.get_loc(stop)

        idx_common = idx_common[start_loc : stop_loc + 1]

    if len(idx_common) == 0:
        raise UserTypeError("The re-indexed time series is empty.")

    # Reindex data to common index
    out = [_reindex(d, idx_common, method, kind, bounded) for d in data]

    return out


@check_types
def _reindex(
    data: pd.Series,
    new_index: pandas.core.indexes.datetimes.DatetimeIndex,
    method: method_options,
    kind: kind_options,
    bounds_error: bool,
) -> pd.Series:
    # Create x values for output time series
    x_uniform = new_index.view(np.int64)

    # extract time series as pd.Series and drop NaNs
    observations = data.dropna()

    # x and y datapoints used to construct linear piecewise function
    x_observed = observations.index.view(np.int64)
    y_observed = observations.values

    if bounds_error:
        fill_value = None
    else:
        fill_value = "extrapolate"

    # interpolator function
    interper = scipy.interpolate.interp1d(
        x_observed, y_observed, kind=method, bounds_error=bounds_error, fill_value=fill_value, assume_sorted=True
    )

    # If pointwise, sample directly from interpolated (or original) points
    if kind == "pointwise":
        y_uniform = interper(x_uniform)
    else:
        y_uniform = functional_mean(interper, x_uniform)

    series = pd.Series(data=y_uniform, index=new_index, name=data.name)

    return series



# Copyright 2022 Cognite AS
from indsl import (
    data_quality,
    detect,
    equipment,
    exceptions,
    filter,
    fluid_dynamics,
    forecast,
    not_listed_operations,
    oil_and_gas,
    regression,
    resample,
    signals,
    smooth,
    statistics,
    ts_utils,
)



# Determine rate of emissions from combustor
from typing import Optional

import pandas as pd

from indsl.resample.resample import resample_to_granularity
from indsl.ts_utils.numerical_calculus import trapezoidal_integration
from indsl.type_check import check_types


@check_types
def emissions_factor_combustor(
    emissions_factor: Optional[float] = None,
    heating_value: Optional[float] = None,
    carbon_content: Optional[float] = None,
) -> float:
    r"""Combustor emissions factor.

    This function calculates the emissions factor of a combustor (e.g., gas turbine, flare, etc.). Three different methods can be used
    to perform this calculation:

    * Method 1: Specify the emissions factor in CO2 emitted per mass or volume unit.
    * Method 2: Specify the emissions factor in CO2 per energy unit and convert it to mass or volume rate using the heating value.

    .. math::
        \mathrm{EF [per\ mass\ or\ volume\ of\ fuel] = EF [per\ unit\ of\ energy] \times Heating\ Value}

    * Method 3: Specify the carbon content of the fuel and use that as the emissions factor.

    Args:
        emissions_factor: CO2 emitted either per mass, volume or energy unit.
        heating_value: This can be either the lower or higher heating value depending on the country's legal requirements.
        carbon_content: Carbon content of fuel (mass of carbon by mass or volume of fuel).

    Returns:
        float: Emissions factor in per mass or volume unit
    """
    # Method 1
    if emissions_factor and (heating_value is None) and (carbon_content is None):
        return emissions_factor

    # Method 2
    elif emissions_factor and heating_value and (carbon_content is None):
        return heating_value * emissions_factor

    # Method 3
    elif (emissions_factor is None) and (heating_value is None) and carbon_content:
        return carbon_content

    raise RuntimeError(
        "Incorrect arguments defined. Please see documentation in detail for how to specify the correct inputs."
    )


# Rate of emissions
@check_types
def rate_of_emissions(data: pd.Series, emissions_factor: float) -> pd.Series:
    r"""Rate of emissions.

    This function calculates the rate of emissions generated by a power consumer or a combustor.

    .. math::
        \mathrm{Rate\ of\ emissions = Emissions\ Factor \times Power\ (or\ Fuel\ Flow\ Rate)}

    The units for the time series and the emissions factor should be consistent to deliver an expected unit for the output in kg (or tonnes) CO2/time.

    Args:
        data: Time series representing either power or rate of fuel consumption.
        emissions_factor: CO2 emitted per unit of energy, mass, or volume.

    Returns:
        pandas.Series: Rate of emissions.
    """
    return data * emissions_factor


@check_types
def cumulative_co2_production(rate_of_emissions: pd.Series, start_date: Optional[pd.Timestamp] = None) -> pd.Series:
    r"""Cumulative CO2 production.

    This function calculates the total CO2 production according to the rate of emissions. The total is calculated by performing trapezoidal integration
    over time (granularity of 1 hour). The rate of emissions is resampled to 1-hour granularity accordingly. If no start time is specified,
    it will default to the start of the current year.

    Args:
        rate_of_emission: Rate of CO2 released over time [mass/time].
        start_date: Start date to begin cumulative calculation.

    Returns:
        pandas.Series: Cumulative CO2 emissions

    """
    # Assign start_date to be start of year if not defined
    if start_date is None:
        start_date = pd.Timestamp(year=pd.Timestamp("now").year, month=1, day=1)

    data = resample_to_granularity(rate_of_emissions.loc[start_date:], granularity=pd.Timedelta("1h"))

    return trapezoidal_integration(data, time_unit=pd.Timedelta("1h"))


@check_types
def cumulative_co2_cost(
    data: pd.Series, co2_cost_factor: float, emissions_factor: float, start_date: Optional[pd.Timestamp] = None
) -> pd.Series:
    r"""Cumulative CO2 cost.

    This function calculates the cumulative cost of CO2 for either a combustor or power consumer. It calculates the rate of CO2 emitted
    and then uses that to calculate total CO2 emitted. This is then multiplied by the cost factor to get the total cumulative cost. Note that the co2_cost_factor, emissions_factor, and data must have consistent units to generate a currency output.

    Args:
        data: Power consumption or fuel consumption.
        co2_cost_factor: Cost per mass of CO2 emitted.
        emissions_factor: Mass of CO2 emitted per mass, volume, or energy of power source consumed.
        start_date: Start date to begin cumulative calculation.

    Returns:
        pandas.Series: Cumulative cost of CO2 emissions
    """
    rate_co2_produced = rate_of_emissions(data, emissions_factor)
    total_co2_produced = cumulative_co2_production(rate_co2_produced, start_date=start_date)

    return total_co2_produced * co2_cost_factor



# Copyright 2022 Cognite AS
from .engineering_calcs import productivity_index
from .gas_density_calcs import calculate_gas_density
from .shut_in_detector import calculate_shutin_interval
from .shut_in_variables import calculate_shutin_variable
from .well_prod_status import calculate_well_prod_status


TOOLBOX_NAME = "Oil and gas"

__all__ = [
    "productivity_index",
    "calculate_shutin_interval",
    "calculate_shutin_variable",
    "calculate_well_prod_status",
    "calculate_gas_density",
]



# Copyright 2021 Cognite AS
import numpy as np
import pandas as pd

from numba import njit, prange
from scipy.integrate import cumulative_trapezoid

from indsl import versioning
from indsl.exceptions import UserValueError
from indsl.type_check import check_types
from indsl.validations import validate_series_has_minimum_length, validate_series_is_not_empty

from . import numerical_calculus_v1  # noqa


@check_types
@versioning.register(version="2.0", changelog="Removed granlularity argument; updated parameter types")
def trapezoidal_integration(series: pd.Series, time_unit: pd.Timedelta = pd.Timedelta("1 h")) -> pd.Series:
    """Integration.

    Cumulative integration using trapezoidal rule with an optional user-defined time unit.

    Args:
        series: Time series.
        time_unit: Frequency.
            User defined granularity to potentially override unit of time. Defaults to 1 h.
            Accepts integer followed by time unit string (ms|s|m|h|d). For example: '1s', '5m', '3h' or '1d'.

    Returns:
        pandas.Series: Cumulative integral.
    """
    validate_series_is_not_empty(series)
    arr = cumulative_trapezoid(series, series.index.view(np.int64) / time_unit.value, initial=0.0)
    return pd.Series(arr, index=series.index)


@check_types
@versioning.register(version="2.0", changelog="Removed granlularity argument; updated parameter types")
def differentiate(series: pd.Series, time_unit: pd.Timedelta = pd.Timedelta("1h")) -> pd.Series:
    """Differentiation.

    Differentiation (finite difference) using a second-order accurate numerical method (central difference).
    Boundary points are computed using a first-order accurate method.

    Args:
        series: Time series.
        time_unit: Frequency.
            User defined granularity to potentially override unit of time. Defaults to 1 h.
            Accepts integer followed by time unit string (ms|s|m|h|d). For example: '1s', '5m', '3h' or '1d'.

    Returns:
        pandas.Series: First order derivative.
    """
    validate_series_has_minimum_length(series, 2)
    arr = np.gradient(series, series.index.view(np.int64) / time_unit.value)
    return pd.Series(arr, index=series.index)


@njit(parallel=True)
def window_index(np_datetime_ns, windowlength_in_ns) -> np.ndarray:
    """Sliding window indexing.

    Returns a np.ndarray where the index corresponds to the starting point of a window,
    the value at the index corresponds to numerical indexes for the end of the window.
    This gives the window span each window is supposed to integrate to.

    Args:
        np_datetime_ns (np.ndarray(dtype:int64)): datetime in integer ns
        windowlength_in_ns (int): the length of the window in ns

    Retruns:
        np.ndarray: indexing of timewindows
    """
    from_to_index = np.zeros(len(np_datetime_ns), dtype=np.int64)
    end = np_datetime_ns[len(np_datetime_ns) - 1]
    for i in prange(len(from_to_index)):
        end_of_window_ns = np_datetime_ns[i] + windowlength_in_ns
        from_to_index[i] = (np.abs(np_datetime_ns - end_of_window_ns)).argmin()  # finds the nearest index
        if end <= np_datetime_ns[i] + windowlength_in_ns:
            break
    return from_to_index


@njit(parallel=True)
def integrate_windows(
    values: np.ndarray, dt: np.ndarray, from_to_index: np.ndarray, number_of_windows: int
) -> np.ndarray:
    """Integrate the windows.

    Performs the integration, through the trapezoidal- or midpoint-method.
    Since all the from and to window indexes are available and all inputs are np.ndarrays, this is parallelized and sped up with numba.

    Args:
        values: Values to integrate corrected to the integrand rate.
        dt: The time in milliseconds between datapoints.
        from_to_index: index is the start of a window and value is the end index of the window.
        number_of_windows: the number of windows to integrate

    Returns:
        numpy.ndarray: The result of each integrated window.
    """
    res = np.empty(number_of_windows, dtype=np.float64)
    for i in prange(number_of_windows):
        cumulative = 0  # reset accumulator for new window (but numba does parallel magic in the background)
        for j in prange(i, from_to_index[i]):
            cumulative += ((values[j - 1] + values[j]) / 2) * dt[j - 1]
        res[i] = cumulative
    return res


@check_types
def sliding_window_integration(
    series: pd.Series,
    window_length: pd.Timedelta = pd.Timedelta("5m"),
    integrand_rate: pd.Timedelta = pd.Timedelta("1h"),
) -> pd.Series:
    """Sliding window integration.

    Siding window integration using trapezoidal rule.

    Args:
        series: Time series.
        window_length: window_length
            the length of time the window. Defaults to '5 minute'. Valid time units are:

                * ‘W’, ‘D’, ‘T’, ‘S’, ‘L’, ‘U’, or ‘N’
                * ‘days’ or ‘day’
                * ‘hours’, ‘hour’, ‘hr’, or ‘h’
                * ‘minutes’, ‘minute’, ‘min’, or ‘m’
                * ‘seconds’, ‘second’, or ‘sec’
                * ‘milliseconds’, ‘millisecond’, ‘millis’, or ‘milli’
                * ‘microseconds’, ‘microsecond’, ‘micros’, or ‘micro’
                * ‘nanoseconds’, ‘nanosecond’, ‘nanos’, ‘nano’, or ‘ns’.
        integrand_rate: integrand_rate.
            if the integrands rate is per sec, per hour, per day.
            Defaults to '1 hour'. Valid time units are:

                * ‘W’, ‘D’, ‘T’, ‘S’, ‘L’, ‘U’, or ‘N’
                * ‘days’ or ‘day’
                * ‘hours’, ‘hour’, ‘hr’, or ‘h’
                * ‘minutes’, ‘minute’, ‘min’, or ‘m’
                * ‘seconds’, ‘second’, or ‘sec’
                * ‘milliseconds’, ‘millisecond’, ‘millis’, or ‘milli’
                * ‘microseconds’, ‘microsecond’, ‘micros’, or ‘micro’
                * ‘nanoseconds’, ‘nanosecond’, ‘nanos’, ‘nano’, or ‘ns’.

    Returns:
        pandas.Series: Time series
    """
    validate_series_is_not_empty(series)

    if max(series.index) - min(series.index) <= window_length:
        raise UserValueError("Integration time window length too long compared to the time window of the dataset.")

    if window_length <= pd.Timedelta("0ms"):
        raise UserValueError("Insert non negative timedelta for window tolength")

    if integrand_rate <= pd.Timedelta("0ms"):
        raise UserValueError("Insert non negative timedelta for integrand rate")

    dt = np.diff(series.index.view(np.int64))  # dt in nanosec
    windowlength_in_ns = int(window_length.total_seconds() * 1e9)
    np_datetime_ns = series.index.view(np.int64)  # points to underlying np.ndarray with timestamps in nanoseconds
    from_to_index = window_index(np_datetime_ns, windowlength_in_ns)

    if from_to_index[0] == 0:
        raise UserValueError("Too small sliding window size, increase window or resample with higher frequency")

    # This code finds the number of windows to integrate
    if np.all(from_to_index != 0):
        n_windows = len(from_to_index) - 1
    else:
        n_windows = np.min(np.where(from_to_index == 0))

    integrand = series.to_numpy() / (integrand_rate.total_seconds() * 1e9)  # corrects the integrand rate to ns
    res = integrate_windows(integrand, dt, from_to_index, n_windows)
    new_index = series.index[
        len(series) - n_windows + 1 :
    ]  # difference in the index due to the first window must be calculated before values come in.
    return pd.Series(res[1:], index=new_index)



from typing import Optional

import numpy as np
import pandas as pd

from ..type_check import check_types


@check_types
def white_noise(
    data: pd.Series,
    snr_db: float = 30,
    seed: Optional[int] = None,
) -> pd.Series:
    """Add white noise.

    Adds white noise to the original data using a given signal-to-noise ratio (SNR).

    Args:
        data: Time series
        snr_db: SNR.
            Signal-to-noise ratio (SNR) in decibels. SNR is a comparison of the level of a signal to the level of
            background noise. SNR is defined as the ratio of signal power to noise power. A ratio higher than 1
            indicates more signal than noise. Defaults to 30.
        seed: Seed.
            A seed (integer number) to initialize the random number generator. If left empty, then a fresh,
            unpredictable value will be generated. If a value is entered, the exact random noise will be generated if
            the time series data and date range are not changed.

    Returns:
        pandas.Series: Output
            Original data plus white noise.
    """
    data_power = data.var()
    # Linear SNR
    snr = 10.0 ** (snr_db / 10.0)
    noise_power = data_power / snr
    rng = np.random.default_rng(seed)
    white_noise = np.sqrt(noise_power) * rng.standard_normal(len(data))

    return data + white_noise



# Copyright 2022 Cognite AS
import pandas as pd

from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index, validate_series_is_not_empty, validate_timedelta_unit


@check_types
def rolling_stddev_timedelta(
    data: pd.Series, time_window: pd.Timedelta = pd.Timedelta(minutes=15), min_periods: int = 1
) -> pd.Series:
    """Rolling stdev of time delta.

    Rolling standard deviation computed for the time deltas of the observations. This
    metric aims to measure the amount of variation or dispersion in the frequency of time series data points.

    Args:
        data: Time series.
        time_window: Time window.
            Length of the time period to compute the standard deviation for. Defaults to 'minutes=15'.
            Time unit should be in days, hours, minutes, or seconds. Accepted formats can be found `here
            (https://pandas.pydata.org/docs/reference/api/pandas.Timedelta.html)`_.
        min_periods: Minimum samples.
            Minimum number of observations required in the given time window (otherwise, the result is set to 0).
            Defaults to 1.

    Returns:
        pandas.Series: Time series

    Raises:
        UserTypeError: data is not a time series
        UserValueError: data is empty
        UserTypeError: time_window is not of type pandas.Timedelta
    """
    validate_series_has_time_index(data)
    validate_series_is_not_empty(data)
    validate_timedelta_unit(time_window)

    return (
        data.index.to_series()
        .diff()
        .astype("timedelta64[ms]")
        .fillna(0)
        .rolling(window=time_window, min_periods=min_periods)
        .std()
        .fillna(0)
    )



# Copyright 2022 Cognite AS
from typing import Union

import pandas as pd

from indsl.type_check import check_types


@check_types
def bands(data: pd.Series, period: str = "1h", K: float = 2.0, as_json: bool = True) -> Union[str, pd.DataFrame]:
    """Confidence bands.

    Confidence bands, also known as Bollinger Bands, are a statistical characterization of a time series fluctuations.
    The confidence bands display a graphical envelope (upper and lower bands) given by the deviation (expressed by the envelope's width). The envelope width is estimated as a factor of the standard deviation for a given
    time period.

    Two input parameters are required to describe the historical behavior of the data, a time window, N, and a
    multiplication factor, K. The window influences the "responsiveness" of the bands to magnitude and frequency of data
    variations. The multiplication factor influences the width of the envelope.

    The Bollinger Bands consist of an N-period moving average (MA) and upper and lower bands at K times an N-period
    standard deviation above and below the moving average (MA +/- K*stdev).

    Args:
        data: Time series.
        period: Window.
            Window length in seconds. Used to estimate the moving average and standard deviation. Defaults to 3600.
        K: Factor.
            Factor used to estimate the width of the envelope K*stdev. Defaults to 2.
        as_json: JSON?
            Return a json dictionary (True) or a pandas DataFrame (False). Defaults to True.

    Returns:
        JSON or pandas.DataFrame: Time index, moving average, and upper and lower rolling confidence bands.
    """
    # TODO: Add the option of entering a float as period in minutes
    # TODO: Implement left-padding to avoid a result with no data (NaN) at the start of the time series

    res = pd.DataFrame(columns=["avg", "lower", "upper"], index=data.index)
    # res.index.names = ["timestamp"]

    res.avg = data.rolling(window=period, closed="right").mean()
    res.upper = res.avg + K * data.rolling(window=period, closed="right").std()
    res.lower = res.avg - K * data.rolling(window=period, closed="right").std()
    res = res.dropna()

    # Convert to json oriented by index
    if as_json:
        res = res.to_json(orient="index")

    return res



# Copyright 2022 Cognite AS
import warnings

from contextlib import suppress
from typing import Literal, Union

import numpy as np
import pandas as pd
import scipy.interpolate

from indsl import versioning
from indsl.ts_utils.ts_utils import functional_mean, is_na_all

from ..exceptions import UserTypeError, UserValueError
from ..type_check import check_types


@versioning.register(version="1.0", deprecated=True)
@check_types
def interpolate(
    data: pd.Series,
    method: Literal["linear", "ffill", "stepwise", "zero", "slinear", "quadratic", "cubic"] = "linear",
    kind: Literal["pointwise", "average"] = "pointwise",
    granularity: str = "1s",
    bounded: int = 1,
    outside_fill: str = None,
) -> Union[pd.DataFrame, pd.Series]:
    """Interpolation.

    Interpolates and resamples data with a uniform sampling frequency.

    Args:
        data: Time series.

        method: Method.
            Specifies the interpolation method. Defaults to "linear". Possible inputs are :

            * 'linear': linear interpolation.
            * 'ffill': forward filling.
            * 'stepwise': yields same result as ffill.
            * 'zero', 'slinear', 'quadratic', 'cubic': spline interpolation of zeroth, first, second or third order.

        kind: Kind.
            Specifies the kind of returned data points. Defaults to "pointwise".  Possible inputs are :

                * 'pointwise': returns the pointwise value of the interpolated function for each timestamp.
                * 'average': returns the average of the interpolated function within each time period.

        granularity: Frequency.
            Sampling frequency or granularity of the output (e.g. '1s' or '2h'). 'min' refers to minutes and 'M' to
            months. Defaults to "1s".

        bounded: Bounded.
            - 1: Fill in for requested points outside of the data range.
            - 0: Ignore said points. Defaults to 1.

            Default behaviour is to raise a UserValueError if the data range is not within start and end and no outside fill
            method is specified (value 1).

        outside_fill: Outside fill.
            Specifies how to fill values outside input data range ('None' or 'extrapolate'). Defaults to None.

    Returns:
        pandas.Series: Interpolated time series.

    Raises:
        UserTypeError: data is not a time series
        UserTypeError: If specified, outside_fill must be either 'None' or 'extrapolate'.
        Warning: Empty data time series
        Warning: All data in timeseries is nan
    """
    # Check for empty time series
    if len(data) == 0:
        warnings.warn("The time series is empty.", RuntimeWarning)
        return data

    # Check if all values are NaN
    if is_na_all(data):
        warnings.warn("All values in the time series are NaN.", RuntimeWarning)
        return data

    # Allow for other ways of defining forward filling for stepwise functions
    method = "previous" if method in ("ffill", "stepwise") else method  # type: ignore

    # check that bounded is correctly specified
    if bounded not in (0, 1):
        raise UserTypeError("If specified, bounded must be either 1 or 0.")
    else:
        bounds_error = bool(bounded)

    # Get outside fill value
    if outside_fill == "None":
        outside_fill = None
    if outside_fill not in (None, "extrapolate") and outside_fill is not None:
        raise UserTypeError("If specified, outside_fill must be either 'None' or 'extrapolate'.")

    # Get start and end dates and store as datetime
    start_dt = _validate(data.index[0])
    end_dt = _validate(data.index[-1])

    # Output timestamps for uniform time series
    timestamps = pd.date_range(start_dt, end_dt, freq=granularity)

    # Create uniform x values for output time series
    x_uniform = np.array([timestamp.timestamp() for timestamp in timestamps])

    # extract time series as pd.Series and drop NaNs
    observations = data.dropna()

    # x and y datapoints used to construct linear piecewise function
    x_observed = np.array([index.timestamp() for index in observations.index])
    y_observed = observations.values.squeeze()

    # interpolator function
    interper = scipy.interpolate.interp1d(
        x_observed, y_observed, kind=method, bounds_error=bounds_error, fill_value=outside_fill
    )

    # If pointwise, sample directly from interpolated (or original) points
    if kind == "pointwise":
        y_uniform = interper(x_uniform)
    elif kind == "average":
        y_uniform = functional_mean(interper, x_uniform)

    series = pd.Series(data=y_uniform, index=timestamps, name=data.name)

    return series


@check_types
def _validate(date_text: Union[str, pd.Timestamp]) -> pd.Timestamp:
    """Validates data format.

    Args:
        date_text (Union[str, pd.Timestamp]): Date in string format.

    Raises:
        UserValueError: If date string format does not match YYYY-MM-DD hh:mm:ss.

    Returns:
        pandas.Timestamp: String converted to pandas.Timestamp.
    """
    with suppress(ValueError):
        return pd.Timestamp(date_text)
    raise UserValueError("Incorrect data format, should be YYYY-MM-DD hh:mm:ss")



# Copyright 2021 Cognite AS
from typing import List

import numpy as np
import pandas as pd

from numpy.polynomial import polynomial
from scipy.stats import t as student_dist

from indsl.filter import trend_extraction_hilbert_transform as get_trend
from indsl.resample import reindex
from indsl.smooth import sg

from ..exceptions import UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_minimum_length


@check_types
def extreme(data: pd.Series, alpha: float = 0.05, bc_relaxation: float = 0.167, poly_order: int = 3) -> pd.Series:
    r"""Extreme outliers removal.

    Outlier detector and removal based on the `paper from Gustavo A. Zarruk
    <https://iopscience.iop.org/article/10.1088/0957-0233/16/10/012/meta>`_. The procedure is as follows:

         * Fit a polynomial curve to the model using all the data
         * Calculate the studentized deleted (or externally studentized) residuals
         * These residuals follow a t distribution with degrees of freedom n - p - 1
         * Bonferroni critical value can be computed using the significance level (alpha) and t distribution
         * Any values that fall outside of the critical value are treated as anomalies

    Use of the hat matrix diagonal allows for the rapid calculation of deleted residuals without having to refit
    the predictor function each time.

    Args:
        data: Time series.
        alpha: Significance level.
            This is a number higher than or equal to 0 and lower than 1. In statistics, the significance level is the
            probability of rejecting the null hypothesis when true. For example, a significance level of 0.05 means that
            there is a 5% risk detecting an outlier that is not a true outlier.
        bc_relaxation: Relaxation factor
            for the Bonferroni critical value. Smaller values will make anomaly detection more conservative. Defaults
            to 1/6.
        poly_order: Polynomial order.
            It represents the order of the polynomial function fitted to the original time series.
            Defaults to 3.

    Returns:
        pandas.Series: Time series.

    Raises:
        UserValueError: Alpha must be a number between 0 and 1
    """
    # Check inputs
    data = data.dropna()

    validate_series_has_minimum_length(data, 3)
    if not 0 <= alpha < 1:
        raise UserValueError("The significance level must be a number higher than or equal to 0 and lower than 1")

    x, y = _split_timeseries_into_time_and_value_arrays(data)

    # Create a polynomial fit and apply the fit to data
    coefs = polynomial.polyfit(x, y, poly_order)
    y_pred = polynomial.polyval(x, coefs)

    hat_diagonal = _calculate_hat_diagonal(x)

    # Calculate degrees of freedom (n-p-1)
    n = len(y)
    dof = n - 3  # Using p = 2 for a model based on a single time series

    # Calculate Bonferroni critical value
    bc = student_dist.ppf(1 - alpha / (2 * n), df=dof) * bc_relaxation

    t_res = _calculate_residuals_and_normalize_them(dof, hat_diagonal, y, y_pred)

    # Return filtered dataframe with the anomalies removed
    mask = np.logical_and(t_res < bc, t_res > -bc)
    return pd.Series(y[mask], index=data.index[mask])


@check_types
def out_of_range(
    data: pd.Series,
    use_sg_pass01: bool = True,
    use_sg_pass02: bool = False,
    window_length: List[int] = [20, 20],
    polyorder: List[int] = [3, 3],
    alpha: List[float] = [0.05, 0.05],
    tolerance: List[float] = [0.08, 0.01],
    bc_relaxation: List[float] = [0.25, 0.5],
    return_outliers: bool = True,
) -> pd.Series:
    r"""Out of range.

    The main objective of this function is to detect data points outside the typical range or  unusually far
    from the main trend. The method is data adaptive; i.e. it  should work independent of the data characteristics. The
    only exception, the method is designed for *non-linear, non-stationary sensor data*, one of the most common type
    of time series in physical processes.

    Outliers are detected using an iterative, data-adaptive method based on digital filtering, signal decomposition and
    statistical inference. Additional details on the analytical methods are provided below. But it is basically a
    three step process, carried out in two iterations. The three steps are::

    1. Estimate the main trend of the time series using one of two methods:

        * Savitzky-Golay smoothing (`SG <https://en.wikipedia.org/wiki/Savitzky%E2%80%93Golay_filter>`_).
        * Empirical mode decomposition - Hilbert Huang Transform (`EMD-HHT <https://emd.readthedocs.io/en/stable/>`_).

    2. Estimate the `studentized residuals <https://en.wikipedia.org/wiki/Studentized_residual>`_.

    3. Identify outliers using the `bonferroni correction <https://mathworld.wolfram.com/BonferroniCorrection.html>`_
       or bonferroni outlier test.

    The results from the first iteration (new time series without the detected outliers) are uses to carry out the
    second iteration. Each of the methods used to estimate the main trend has pros and cons, but their combination is
    an effective way to leverage each of their strengths. The `Student's t-distribution
    <https://en.wikipedia.org/wiki/Student%27s_t-distribution>`_ is used as it useful when
    estimating the mean of a normally distributed population in situations where the sample size is small and
    the population's standard deviation is unknown. Finally, the bonferroni correction or Boneferroni test is
    a simple and efficient method to tests for extreme outliers. Additional details on each of these methods is
    provided below.

    **Savitzky-Golay Smoother**:

    The SG smoother is a digital filter ideal for smoothing data without distorting the data tendency. But its true
    value comes from being independent of the sampling frequency (unlike most digital filters). Hence, simple
    and robust to apply on data with non-uniform sampling (i.e. industrial sensor data). Two parameters are required
    to apply the SG smoother, a point-wise `window length` and a `polynomial order`. The `window length` is the number
    of data points used to estimate the local trend and the `polynomial order` is used to fit those data points with a
    linear (order 1) or non-linear (order higher than 1) fit.

    **Empirical Mode Decomposition - Hilbert Huang Transform**:

    This is a data-adaptive method that decomposes a signal into oscillatory components and rejects those with low
    energy (e.g.. high frequency random noise) base on given energy `tolerance` or threshold. A small tolerance
    will allow more oscillatory components as  part of the main (more high-frequency components), while a large
    `tolerance` will result on a smoother trend. The algorithm implemented here is an adaptation from the
    `EMD-HHT <https://emd.readthedocs.io/en/stable/>`_ library. For more information on the EMD-HHT method
    consult the original paper in the Journal of Open Source Software
    `JOSS <https://joss.theoj.org/papers/10.21105/joss.02977>`_.

    **Studentized residuals and the Bonferroni Correction**:

    The `Student's t-distribution <https://en.wikipedia.org/wiki/Student%27s_t-distribution>`_ is typically used when
    the sample size is small and the standard deviation is
    unknown. In this case we assume that there will be a small amount of out of range values and the statistical
    properties of the data are not known. By studentizing the residuals is analogous to normalizing the data and it
    a useful technique for detecting outliers. Furthermore, we apply the bonferroni correction to test if a
    residual is an outlier or not. To studentize the residual an `significance lebel` must be defined. Furthermore, the
    Bonferroni Correction is very conservative before it classifies a data point as an outlier. Consequently, a
    `sensitivity` factor is used to relax the test and identifying points that are located close to the main trend
    and that can be removed in the second iteration.

    Args:
        data: Time series.
        use_sg_pass01: Pass 1: Simple smoother.
            Use the Savitzky-Golay (SG) smoother in the first iteration to determine the main trend of the signal.
            If set to `True`, the parameters `window length` and `polynomial order` are required. If set to `False`,
            the parameter `tolerance` is required to obtain the trend using the EMD-HHT
            method. The default is `True`.
        use_sg_pass02: Pass 2: Simple smoother.
            Use the Savitzky-Golay (SG) smoother in the second iteration to determine the main trend of the signal.
            If set to `True`, the parameters `window length` and `polynomial order` are required. If set to `False`,
            the parameter `tolerance` is required to obtain the trend using the EMD-HHT
            method. The default is `False`.
        window_length: Window length.
            Point-wise window length used to estimate the local trend using the SG smoother. Two integer values are
            required, one for each iteration. If the SG smoother is not used, these values are ignored. Default value
            is 20 data points for both iterations.
        polyorder: Polynomial order.
            It represents the order of the polynomial function fitted to the data when using the SG smoother.
            Default value is 3 data points for both iterations.
        alpha: Significance level.
            Number higher than or equal to 0 and lower than 1. Statistically speaking, the significance level
            is the probability of detecting an outlier that is not a true outlier. A value of 0.05 means that
            there is a 5% risk of detecting an outlier that is not a true outlier. Defaults to 0.05 for both
            iterations.
        tolerance: Tolerance.
            Energy tolerance (or threshold) used to validate if an oscillatory component identified by the EMD-HHT
            methodology should be included as part of the main trend.
            Defaults to 0.08 and 0.01.
        bc_relaxation: Sensitivity.
            Number higher than 0 used to make outlier detection more or less sensitive. Smaller values will make
            the detection more conservative. Defaults 0.25 and 0.5 for the first and second iterations, respectively.
        return_outliers: Output outliers.
            If selected (`True`) the method outputs the detected outliers, otherwise the filtered (no-outliers)
            time series is returned. Defaults to `True`.

    Returns:
        pandas.Series: Time series.
    """
    # Check inputs
    data = data.dropna()

    validate_series_has_minimum_length(data, 3)
    _validations(alpha, bc_relaxation, tolerance, window_length, polyorder)

    x, y = _split_timeseries_into_time_and_value_arrays(data)

    # ==========
    # FIRST PASS
    # ==========

    # Estimate the trend using Savitzky-Golay smoother or Empirical Mode Decomposition (EMD) and the Hilbert Huang Transform (HHT)
    if use_sg_pass01:
        trend_pass01 = sg(data, window_length=window_length[0], polyorder=polyorder[0])
    else:
        trend_pass01 = get_trend(data, error_tolerance=tolerance[0])

    # Detect outliers based on first trend estimate
    x, y = _split_timeseries_into_time_and_value_arrays(data)
    y_pred_pass01 = trend_pass01.to_numpy()
    hat_diagonal = _calculate_hat_diagonal(x)

    # Calculate degrees of freedom (n-p-1)
    n = len(y)
    dof = n - 3  # Using p = 2 for a model based on a single time series

    # Calculate Bonferroni critical value and studentized residuals
    bc = student_dist.ppf(1 - alpha[0] / (2 * n), df=dof) * bc_relaxation[0]
    t_res = _calculate_residuals_and_normalize_them(dof, hat_diagonal, y, y_pred_pass01)

    # Boolean mask where outliers are detected
    mask = np.logical_and(t_res < bc, t_res > -bc)
    filtered_ts_pass01 = pd.Series(y[mask], index=data.index[mask])  # Remove detected outliers from time series

    # ===========
    # SECOND PASS
    # ===========

    # Filtering parameters
    if use_sg_pass02:
        trend_pass02 = sg(filtered_ts_pass01, window_length=window_length[1], polyorder=polyorder[1])
    else:
        trend_pass02 = get_trend(filtered_ts_pass01, error_tolerance=tolerance[1])

    y_pred_pass02 = reindex(trend_pass02, data)
    y_pred_pass02 = y_pred_pass02.to_numpy()

    bc_pass02 = student_dist.ppf(1 - alpha[1] / (2 * n), df=dof) * bc_relaxation[1]
    t_res_pass02 = _calculate_residuals_and_normalize_them(dof, hat_diagonal, y, y_pred_pass02)
    #
    # Boolean mask where outliers are detected
    mask_pass02 = np.logical_and(t_res_pass02 < bc_pass02, t_res_pass02 > -bc_pass02)
    filtered_ts_pass02 = pd.Series(
        y[mask_pass02], index=data.index[mask_pass02]
    )  # Remove detected outliers from time series
    outliers_pass02 = pd.Series(y[~mask_pass02], index=data.index[~mask_pass02])

    # Assign result selected in the arguments
    if return_outliers:
        res = outliers_pass02
    else:
        res = filtered_ts_pass02

    return res


def _validations(alpha, bc_relaxation, tolerance, window_length, polyorder):
    if not all(i > 0.0 for i in tolerance):
        raise UserValueError("The Error Tolerance must be a number higher than zero.")
    if not all(0 <= i < 1 for i in alpha):
        raise UserValueError(
            "The Significance Level (alpha) must be a number higher than or equal to 0 and lower " "than 1"
        )
    if not all(i > 0 for i in bc_relaxation) > 0:
        raise UserValueError("The Relaxation Factor must be a number higher than 0")
    if len(window_length) != 2:
        raise UserValueError(f"The window length requires two values, got {window_length}")
    if len(polyorder) != 2:
        raise UserValueError(f"The polynomial order parameter requires two values, got {polyorder}")
    if len(alpha) != 2:
        raise UserValueError(f"The Significance Level (alpha) parameter requires two values, got {alpha}")
    if len(tolerance) != 2:
        raise UserValueError(f"The tolerance order parameter requires two values, got {tolerance}")
    if len(bc_relaxation) != 2:
        raise UserValueError(f"The sensitivity parameter requires two values, got {bc_relaxation}")


def _split_timeseries_into_time_and_value_arrays(data: pd.Series) -> tuple:
    """Split ts into time and value arrays.

    Method that takes a time series with a datetime index and splits it into two numpy arrays. One with the datetime
    index converted to integer values, starting from zero, and the other array contains the values of the time series.

    Args:
        data (pandas.Seres): Time series with a datetime index

    Returns:
        tuple: Tuple containing two 1-D numpy arrays:
            x -> Datetime index converted to integers starting at 0
            y -> Time series values
    """
    x = (np.array(data.index, dtype=np.int64) - data.index[0].value) / 1e9
    y = data.to_numpy()  # Just to please pandas devs
    return x, y


def _calculate_residuals_and_normalize_them(
    dof: int, hat_diagonal: np.ndarray, y: np.ndarray, y_pred: np.ndarray
) -> np.ndarray:
    """Calculate residuals and normalize.

    Calculate studentized residuals between a signal and fitted (e.g. regression) values.

    Args:
        dof (int): Degrees of freedom of the Student's t-distribution
        hat_diagonal (np.ndarray): Diagonal of the hat matrix
        y (np.ndarray) : Original data points
        y_pred (np.ndarray) : Fitted data points

    Returns:
        numpy.ndarray: Studentized residuals
    """
    res = y - y_pred
    sse = np.sum(res**2)
    t_res = res * np.sqrt(dof / (sse * (1 - hat_diagonal) - res**2))
    return t_res


def _calculate_hat_diagonal(x: np.ndarray) -> np.ndarray:
    """Hat diagonal.

    Generates the hat matrix from 1-D numpy array

    Args:
        x (numpy.ndarray): 1-D array

    Returns:
        numpy.ndarray: hat diagonal

    """
    X_mat = np.vstack((np.ones_like(x), x)).T
    X_hat = X_mat @ np.linalg.inv(X_mat.T @ X_mat) @ X_mat.T

    return X_hat.diagonal()



from typing import Union

import pandas as pd

from .exceptions import UserTypeError, UserValueError


def validate_series_has_time_index(data: pd.Series) -> None:
    """Helper method to validate if provided pandas.Series is of type pandas.DatetimeIndex."""
    if not isinstance(data.index, pd.DatetimeIndex):
        raise UserTypeError(f"Expected a time series, got index type {data.index.dtype}")


def validate_series_is_not_empty(series: Union[pd.Series, pd.DataFrame]) -> None:
    """Helper method to validate if provided pandas.Series has more than 0 values."""
    if len(series) == 0:
        raise UserValueError("Time series is empty.")


def validate_series_has_minimum_length(series: pd.Series, min_len: int) -> None:
    """Helper method to validate if provided pandas.Series has the minimum length specified."""
    if len(series) < min_len:
        raise UserValueError(f"Expected series with length >= {min_len}, got length {len(series)}")


def validate_timedelta_unit(timedelta: pd.Timedelta) -> None:
    """Helper method to validate if the provided pd.Timedelta is larger or equal to 1 second."""
    if timedelta < pd.Timedelta(seconds=1):
        raise UserValueError("Unit of timedelta should be in days, hours, minutes or seconds")



# Copyright 2022 Cognite AS

from .gaps_identification import (
    gaps_identification_iqr,
    gaps_identification_modified_z_scores,
    gaps_identification_threshold,
    gaps_identification_z_scores,
)
from .low_density_identification import (
    low_density_identification_iqr,
    low_density_identification_modified_z_scores,
    low_density_identification_threshold,
    low_density_identification_z_scores,
)
from .negative_running_hours import negative_running_hours_check
from .outliers import extreme, out_of_range
from .rolling_stddev import rolling_stddev_timedelta
from .uncertainty import uncertainty_rstd


TOOLBOX_NAME = "Data quality"

__all__ = [
    "gaps_identification_z_scores",
    "gaps_identification_modified_z_scores",
    "gaps_identification_iqr",
    "gaps_identification_threshold",
    "low_density_identification_iqr",
    "low_density_identification_modified_z_scores",
    "low_density_identification_threshold",
    "low_density_identification_z_scores",
    "extreme",
    "out_of_range",
    "negative_running_hours_check",
    "uncertainty_rstd",
    "rolling_stddev_timedelta",
]



import json
import os

from typing import Literal

import numpy as np
import pandas as pd

from scipy import interpolate as intp

from indsl.type_check import check_types


@check_types
def _error_handling(pvt_table: np.ndarray, pressure: pd.Series, temperature: pd.Series) -> None:
    max_pressure, min_pressure = max(pvt_table[:, [0]])[0], min(pvt_table[:, [0]])[0]
    max_temperature, min_temperature = max(pvt_table[:, [1]])[0], min(pvt_table[:, [1]])[0]

    errors = []
    if max(pressure) > max_pressure:
        errors.append(
            f"max pressure of the sensor is outside the window of the pvt table, max pressure PVT table = {max_pressure}, max pressure sensor = {max(pressure)}"
        )
    if min(pressure) < min_pressure:
        errors.append(
            f"min pressure of the sensor is outside the window of the pvt table, min pressure PVT table = {min_pressure}, min pressure sensor = {min(pressure)}"
        )
    if max(temperature) > max_temperature:
        errors.append(
            f"max temperature of the sensor is outside the window of the pvt table, max temperature PVT table = {max_temperature}, max temperature sensor = {max(temperature)}"
        )
    if min(temperature) < min_temperature:
        errors.append(
            f"min temperature of the sensor is outside the window of the pvt table, min temperature PVT table = {min_temperature}, min temperature sensor = {min(temperature)}"
        )

    if errors:
        msg = "\n           ".join(errors)
        raise Exception(msg)


@check_types
def retrieve_fluid_properties(
    pressure: pd.Series,
    temperature: pd.Series,
    pvt_data: pd.DataFrame,
    param: str,
    interp_method: Literal["linear", "nearest", "cubic"] = "linear",
) -> pd.Series:
    """Retrieve fluid properties.

    This function obtains the value for the selected fluid property (`param`) corresponding to the pressure/temp pairs input as series. The input fluid file type is a .tab file that is an output
    of an equation of state simulator (i.e PVTSim, MultiFlash) and is an input file specifically for OLGA. The 31 fluid property parameters included in the .tab file are the options as an output for this function.
    Note that pressure and temperature sensors have to come from the same location.

    Args:
        pressure: Pressure time series.
        temperature: Temperature time series.
        pvt_data: PVT data.
            This is a PVT table that has been parsed out from a .tab file which is a standardized generated output of
            MultiFlash and PVTSim.
        param: Fluid parameter of interest.
            Fluid property to be estimated.
        interp_method: Method.
            The method used to interpolate the pvt table to obtain fluid properties of a given pressure and temperature.

    Returns:
        pandas.Series: Time series
            Fluid property of selected fluid property ("param") for corresponding pressure and temperature sensors.
            There are 31 fluid properties currently available to output from the fluid file.

    Raises:
        AssertionError: If data in sequence is not from a .tab file.
        AssertionError: If `param` input is not one of the 31 fluid properties available.
        Exception: If the range of the pressure and temperature sensor is outside the pvt table window, the function
        cannot interpolate for fluid property parameters.
    """
    assert (pvt_data.attrs["file_type"]) == "tab", "File is not a '.tab' file"

    column_names, pvt_table = pvt_data.columns, pvt_data.values
    assert param in column_names, (
        "Parameter is not a property of the fluid file. Valid inputs are %s" % column_names.values
    )

    pvt_table_dict_file = os.path.join(os.path.dirname(__file__), "tab_fp_identifier.json")
    pvt_table_dict = json.load(open(pvt_table_dict_file, "rb"))
    description = pvt_table_dict[param]["description"]
    unit = pvt_table_dict[param]["unit"]

    param_idx = [idx for idx, val in enumerate(column_names) if param == val][0]

    _error_handling(pvt_table, pressure, temperature)

    return pd.Series(
        intp.griddata(pvt_table[:, [0, 1]], pvt_table[:, param_idx], (pressure, temperature), method=interp_method),
        index=pressure.index,
        name=f"{description} [{unit}]",
    )



# Copyright 2022 Cognite AS
from .logarithmic_functions import exp, log, log2, log10, logn
from .numerical_calculus import differentiate, sliding_window_integration, trapezoidal_integration
from .operators import absolute, add, arithmetic_mean, div, inv, mod, mul, neg, power, sqrt, sub
from .trigonometric_functions import (
    arccos,
    arccosh,
    arcsin,
    arcsinh,
    arctan,
    arctan2,
    arctanh,
    cos,
    cosh,
    deg2rad,
    rad2deg,
    sin,
    sinh,
    tan,
    tanh,
)
from .utility_functions import (
    bin_map,
    ceil,
    clip,
    floor,
    get_timestamps,
    maximum,
    minimum,
    remove,
    replace,
    round,
    set_timestamps,
    sign,
    threshold,
    time_shift,
    union,
)


TOOLBOX_NAME = "Operators"

__all__ = [
    "add",
    "sub",
    "mul",
    "div",
    "absolute",
    "inv",
    "mod",
    "neg",
    "power",
    "sqrt",
    "sin",
    "cos",
    "tan",
    "arccos",
    "arcsin",
    "arctan",
    "arctan2",
    "deg2rad",
    "rad2deg",
    "sinh",
    "cosh",
    "tanh",
    "arcsinh",
    "arccosh",
    "arctanh",
    "exp",
    "log",
    "log2",
    "log10",
    "logn",
    "clip",
    "replace",
    "remove",
    "maximum",
    "minimum",
    "union",
    "bin_map",
    "get_timestamps",
    "set_timestamps",
    "time_shift",
    "ceil",
    "floor",
    "round",
    "sign",
    "trapezoidal_integration",
    "differentiate",
    "threshold",
    "sliding_window_integration",
    "arithmetic_mean",
]



# Copyright 2022 Cognite AS
import numba
import numpy as np
import pandas as pd

from indsl.exceptions import UserValueError
from indsl.type_check import check_types
from indsl.validations import validate_series_has_time_index, validate_series_is_not_empty


@check_types
def _validate_time_series(x: pd.Series) -> None:
    validate_series_has_time_index(x)
    validate_series_is_not_empty(x)

    if not x.index.is_monotonic_increasing:
        raise UserValueError("Time series index is not increasing.")


@check_types
def _validate_threshold(threshold: float) -> None:
    if threshold < 0:
        raise UserValueError("Threshold should be a non-negative float.")


@numba.njit
def _prolong_indicator(x_value: np.array, x_indicator: np.array) -> np.array:
    """Indicator adjustment.

    Function that prolongs the indicator value for bad data quality - indicator will be set to 1 for as long as data doesn't go back to "normal" (ie., back to the value it had before the value decreased).

    Args:
         x_value: Numpy array with time series values
         x_indicator: Numpy array (with values 0 and 1) indicating bad data quality for given time series x

    Returns:
         np.array: Indicator time series
    """
    i, n = 0, len(x_value)

    while i < n - 1:
        if x_indicator[i] == 1:
            j = i + 1
            while j < n and x_value[i - 1] > x_value[j]:
                x_indicator[j] = 1
                j += 1
            i = j
        else:
            i += 1

    return x_indicator


@check_types
def negative_running_hours_check(x: pd.Series, threshold: float = 0.0) -> pd.Series:
    """Negative running hours.

    The negative running hours model is created in order to automate data quality check for time series with values that
    shouldn't be decreasing over time. One example would be Running Hours (or Hour Count) time series - a specific type
    of time series that is counting the number of running hours in a pump. Given that we expect the number of running
    hours to either stay the same (if the pump is not running) or increase with time (if the pump is running), the
    decrease in running hours value indicates bad data quality. Although the algorithm is originally created for
    Running Hours time series, it can be applied to all time series where the decrease in value is a sign of bad data
    quality.

    Args:
        x: Time series
        threshold: Threshold for value drop.
            This threshold indicates for how many hours the time series value needs to drop (in hours) before we
            consider it bad data quality. The threshold must be a non-negative float. By default, the threshold is set to 0.

    Returns:
        pandas.Series: Time series
            The returned time series is an indicator function that is 1 where there is a decrease in time series
            value, and 0 otherwise. The indicator will be set to 1 until the data gets "back to normal" (that is,
            until time series reaches the value it had before the value drop).


    Raises:
        UserTypeError: x is not a time series
        UserValueError: x is empty
        UserTypeError: index of x is not a datetime
        UserValueError: index of x is not increasing
        UserTypeError: threshold is not a number
        UserValueError: threshold is a negative number
    """
    _validate_time_series(x)

    _validate_threshold(threshold)

    if len(x) < 2:
        return pd.Series([0] * len(x), index=x.index)

    x_diff = x.diff(1)
    x_indicator = (x_diff < -threshold).astype(np.uint8)

    return pd.Series(
        _prolong_indicator(
            x.to_numpy(),
            x_indicator.to_numpy(),
        ).astype(np.int64),
        index=x.index,
    )



import warnings

from typing import Any, List, Literal, Optional, Union

import numpy as np
import pandas as pd

from ..exceptions import UserValueError
from ..type_check import check_types
from ..validations import validate_series_has_time_index


@check_types
def line(
    start_date: Optional[pd.Timestamp] = None,
    end_date: Optional[pd.Timestamp] = None,
    sample_freq: pd.Timedelta = pd.Timedelta("1 m"),
    slope: float = 0,
    intercept: float = 0,
) -> pd.Series:
    """Line.

    Generate a synthetic time series using the line equation. If no end and/or start dates are given, the default
    signal duration is set to 1 day. If no dates are provided, the end date is set to the current date and time.

    Args:
        start_date: Start date.
            The start date of the time series entered as a string, for example: "1975-05-09 20:09:10", or
            "1975-05-09".
        end_date: End date.
            The end date of the time series entered as a string, for example: "1975-05-09 20:09:10", or
            "1975-05-09".
        sample_freq: Frequency.
            Sampling frequency as a time delta, value and time unit. Defaults to '1 minute'. Valid time units are:

                * ‘W’, ‘D’, ‘T’, ‘S’, ‘L’, ‘U’, or ‘N’
                * ‘days’ or ‘day’
                * ‘hours’, ‘hour’, ‘hr’, or ‘h’
                * ‘minutes’, ‘minute’, ‘min’, or ‘m’
                * ‘seconds’, ‘second’, or ‘sec’
                * ‘milliseconds’, ‘millisecond’, ‘millis’, or ‘milli’
                * ‘microseconds’, ‘microsecond’, ‘micros’, or ‘micro’
                * ‘nanoseconds’, ‘nanosecond’, ‘nanos’, ‘nano’, or ‘ns’.
        slope: Slope.
            Line slope. Defaults to 0 (horizontal line).
        intercept: Intercept.
             Y-intercept. Defaults to 0.

    Returns:
        pandas.Series: Time series
            Synthetic time series for a line

    """
    idx = _make_index(start=start_date, end=end_date, freq=sample_freq)
    line_data = _time_array(idx) * slope + intercept

    return pd.Series(data=line_data, index=idx, name="Line")


@check_types
def const_value(value: float = 0, timedelta: pd.Timedelta = pd.Timedelta("1 W")) -> pd.Series:
    """Constant value.

    This function generates a horizontal line. The assumptions when generating the horizontal line
    are that the start date is set as "1970-01-01", the end date is set as "now", and the sampling is "1 week".
    If the number of data points generated exceeds 100000, then the start date is moved forward, such that the
    number of data points generated is not greater than 100000 with "1 week" sampling resolution.

    Args:
        value: Value.
            value. Defaults to 0.
        timedelta: Granularity.
            Sampling frequency as a time delta, value, and time unit. Defaults to one week ('1 W'). Valid time units are:

                * ‘W’, ‘D’, ‘T’, ‘S’, ‘L’, ‘U’, or ‘N’
                * ‘days’ or ‘day’
                * ‘hours’, ‘hour’, ‘hr’, or ‘h’
                * ‘minutes’, ‘minute’, ‘min’, or ‘m’
                * ‘seconds’, ‘second’, or ‘sec’
                * ‘milliseconds’, ‘millisecond’, ‘millis’, or ‘milli’
                * ‘microseconds’, ‘microsecond’, ‘micros’, or ‘micro’
                * ‘nanoseconds’, ‘nanosecond’, ‘nanos’, ‘nano’, or ‘ns’.

    Returns:
        pandas.Series: Time series.
            Synthetic time series for a line

    """
    slope = 0
    idx = _make_index(start="1970-01-01", end="now", freq=timedelta)
    if len(idx) > 1e8:
        idx = idx[len(idx) - 100000 :]

    line_data = _time_array(idx) * slope + value
    return pd.Series(data=line_data, index=idx, name="Constant value")


@check_types
def sine_wave(
    start_date: Optional[pd.Timestamp] = None,
    end_date: Optional[pd.Timestamp] = None,
    sample_freq: pd.Timedelta = pd.Timedelta("1 s"),
    wave_period: pd.Timedelta = pd.Timedelta("1 h"),
    wave_mean: float = 0,
    wave_amplitude: float = 1,
    wave_phase: float = 0,
) -> pd.Series:
    """Sine wave.

    Generate a time series for a `sine wave <https://en.wikipedia.org/wiki/Sine_wave>`_ with a given wave period,
    amplitude, phase, and mean value. If no end and/or start dates are given, the default signal duration is set to
    1 day. If no dates are provided, the end date is set to the current date and time.

    Args:
        start_date: Start date
            Date-time string when the time series starts. The date must be a string, for example:
            "1975-05-09 20:09:10".
        end_date: End date
            Date-time string when the time series starts. The date must be a string, for example:
            "1975-05-09 20:09:10".
        sample_freq: Frequency.
            Sampling frequency as a time delta, value, and time unit. Defaults to '1 minute'. Valid time units are:

                * ‘W’, ‘D’, ‘T’, ‘S’, ‘L’, ‘U’, or ‘N’
                * ‘days’ or ‘day’
                * ‘hours’, ‘hour’, ‘hr’, or ‘h’
                * ‘minutes’, ‘minute’, ‘min’, or ‘m’
                * ‘seconds’, ‘second’, or ‘sec’
                * ‘milliseconds’, ‘millisecond’, ‘millis’, or ‘milli’
                * ‘microseconds’, ‘microsecond’, ‘micros’, or ‘micro’
                * ‘nanoseconds’, ‘nanosecond’, ‘nanos’, ‘nano’, or ‘ns’.
        wave_period: Wave period.
            The time it takes for two successive crests (one wavelength) to pass a specified point. For example, defining
            a wave period of :math:`10 min` will generate one full wave every 10 minutes. The period can not be 0. If
            no value is provided, it is 1 minute. Valid time units are:

                * ‘W’, ‘D’, ‘T’, ‘S’, ‘L’, ‘U’, or ‘N’
                * ‘days’ or ‘day’
                * ‘hours’, ‘hour’, ‘hr’, or ‘h’
                * ‘minutes’, ‘minute’, ‘min’, or ‘m’
                * ‘seconds’, ‘second’, or ‘sec’
                * ‘milliseconds’, ‘millisecond’, ‘millis’, or ‘milli’
                * ‘microseconds’, ‘microsecond’, ‘micros’, or ‘micro’
                * ‘nanoseconds’, ‘nanosecond’, ‘nanos’, ‘nano’, or ‘ns’.
        wave_mean: Mean.
            The wave's mean value. Defaults to 0.
        wave_amplitude: Peak amplitude.
            Maximum absolute deviation from the mean. Defaults to 1.
        wave_phase: Phase.
            Specifies (in radians) where in its cycle the oscillation is at time = 0. When the phase is non-zero, the
            wave is shifted in time. A negative value represents a delay, and a positive value represents an advance.
            Defualts to 0.

    Returns:
        pandas.Series: Sine wave

    """
    if sample_freq.total_seconds() <= 0:
        raise UserValueError(
            f"The sampling frequency must be a value higher than zero. " f"Value provided was {sample_freq}"
        )
    if wave_period.total_seconds() == 0:
        raise UserValueError("The wave period can not be zero")

    idx = _make_index(start=start_date, end=end_date, freq=sample_freq)

    wave_period = wave_period / pd.Timedelta(1, unit="s")

    angular_freq = 2 * np.pi * (1 / wave_period)
    wave_data = wave_amplitude * np.sin(angular_freq * _time_array(idx) + wave_phase) + wave_mean

    return pd.Series(data=wave_data, index=idx, name="Sine wave")


@check_types
def perturb_timestamp(data: pd.Series, magnitude: float = 1) -> pd.Series:
    """Perturb timestamp.

    Perturb the date-time index (timestamp) of the original time series using a normal (Gaussian) distribution
    with a mean of zero and a given standard deviation (magnitude) in seconds.

    Args:
        data: Time series
        magnitude: Magnitude.
            Time delta perturbation magnitude in seconds. Has to be large than 0. Defaults to 1.

    Returns:
        pandas.Series: Time series
            Original signal with a non-uniform time stamp.

    Raises:
        UserTypeError: Only time series with a DateTimeIndex are supported
        UserTypeError: If "magnitude" is not a float
        UserValueError: If "magnitude" is not larger than 0

    """
    validate_series_has_time_index(data)
    if magnitude <= 0:
        raise UserValueError(f"Expected magnitude to be > 0, got {magnitude}")

    time = _time_array(data.index)

    rng = np.random.default_rng(1975)
    time_perturbations = rng.normal(0, magnitude, len(time))
    ts = data.copy()
    ts.index += pd.to_timedelta(time_perturbations, "sec")
    ts.index = ts.index.sort_values()
    # TODO: Add example for Gallery of Charts

    # TODO: Sughestion from Hakon Treider. Another possibility is to do a cumulative sum on time_perturbations to make
    # the jumps "more continuous" / "higher temporal locality".
    return ts.sort_index()


@check_types
def insert_data_gaps(
    data: pd.Series,
    fraction: float = 0.25,
    num_gaps: Optional[int] = None,
    data_buffer: int = 5,
    method: Literal["Random", "Single", "Multiple"] = "Random",
) -> pd.Series:
    """Insert data gaps.

    Method to synthetically remove data, i.e., generate data gaps in a time series. The amount of data points removed
    is defined by the given 'fraction' relative to the original time series.

    Args:
        data: Time series
        fraction: Remove fraction.
            Fraction of data points to remove relative to the original time series. Must be a number higher than 0 and
            lower than 1 (0 < keep < 1). Defaults to 0.25.
        num_gaps: Number of gaps.
            Number of gaps to generate. Only needs to be provided when using the "Multiple" gaps method.
        data_buffer: Buffer.
            Minimum of data points to keep between data gaps and at the start and end of the time series. If the buffer
            of data points is higher than 1% of the number of data points in the time series, the end and start buffer
            is set to 1% of the total available data points.
        method: Method
            This function offers multiple methods to generate data gaps:

                * Random: Removes data points at random locations so that the output time series size is a given
                  fraction  ('Remove fraction') of the original time series. The first and last data points are never
                  deleted. No buffer is set between gaps, only for the start and end of the time series.
                  If the buffer of data points is higher than 1% of the number of data points in the time
                  series, the end and start buffer is set to 1% of the total available data points.
                * Single: Remove consecutive data points at a single location. Buffer data points at the start
                  and end of the time series is kept to prevent removing the start and end of the time series. The
                  buffer is set to the maximum value between 5 data points or 1% of the data points in the signal.
                * Multiple: Insert multiple non-overlapping data gaps at random dates and of random
                  sizes such that the given fraction of data is removed. If the number of gaps is not defined or is
                  less than 2, the function defaults to 2 gaps. To avoid gap overlapping, a minimum of 5 data points are imposed at the signal's start and end and between gaps.

    Returns:
        pandas.Series: Output
            Original time series with synthetically generated data gap(s).

    Raises:
        UserTypeError: data is not a time series
        UserTypeError: fraction is not a number
        UserTypeError: fraction is not a number

    """
    original_length = len(data)
    rng = np.random.default_rng(1975)
    points_to_remove = int(original_length * fraction) + 1
    # 1% Data buffers locations to avoid inserting large data gaps at the start and end of the time series
    buffer1 = max(data_buffer, int(original_length * 0.01))
    if int(original_length * fraction) > original_length - buffer1 * 2 or original_length < 3:
        raise UserValueError(
            f"Not enough data in the time series. The original time series minus the data buffers has "
            f"{original_length} data points. After removing {fraction * 100}% of the data, the "
            f"time series would have {int(original_length * (1 - fraction)) - buffer1 * 2} data points. Please use"
            f"a longer time series or reduce the fraction of data to remove"
        )
    buffer2 = original_length - buffer1
    if method == "Random":
        gap_loc = rng.choice(np.arange(1, original_length - 1), size=points_to_remove, replace=False, shuffle=False)
        return data.loc[data.index.difference(data.index[gap_loc])]
    elif method == "Single":
        gap_loc = int(rng.integers(low=buffer1, high=buffer2, size=1))
        # Move the start of the gap location away from the end of the time series so that the data gap ends right
        # at the buffer
        if gap_loc + points_to_remove > buffer2:
            gap_loc = buffer2 - points_to_remove

        idx = data.index[gap_loc : gap_loc + points_to_remove]
        return data.drop(labels=idx)
    elif method == "Multiple":
        if num_gaps is None or num_gaps < 2:
            num_gaps = 3
        if points_to_remove >= original_length - buffer1 * (num_gaps + 1):
            raise UserValueError(
                "The amount of data to remove exceeds the amount of data point available for creating gaps. Please "
                "reduce the fraction of data to remove or use a time series with more data points"
            )

        # Generate random gap location, sizes and loc ranges that amount to the fraction of data to be removed
        gap_loc = np.sort(rng.integers(low=buffer1, high=buffer2, size=num_gaps))
        gap_size = _random_list_of_integers_that_gives_exact_sum(list_length=num_gaps, sum_output=points_to_remove)
        gap_ranges = list(
            zip(gap_loc, gap_loc + gap_size - 1)
        )  # BE CAREFUL WITH EDGE RANGE INCLUDING/EXCLUDING LAST DATA POINT

        gap_ranges = _handle_overlapping_gaps(gap_ranges=gap_ranges, data_length=len(data))
        idx_loc: List[np.array] = []
        for ii in range(len(gap_ranges)):
            idx_loc = np.append(idx_loc, np.arange(gap_ranges[ii][0], gap_ranges[ii][1] + 1, dtype=int)).astype(int)
        idx = data.index[idx_loc]
        return data.drop(labels=idx)


@check_types
def _handle_overlapping_gaps(gap_ranges: np.array, data_length: int, buffer: int = 5) -> np.ndarray:
    """Handle overlapping gaps.

    Internal method to handle overlapping gaps. The method calculates the
    number of available space to rearrange gaps in such a way that they do not
    overlap. Then traverses the ranges according to their location from left to
    right (time 0 to end) and estimates how many steps ranges to the left or
    right must be moved to comply with gap generation conditions:

    * The first and last 5 data points of the time series must be preserved
    * A minimum number of data points, defined by `buffer` must exist between adjacent gaps
    * The number of available data point to move ranges around can not exceed the number of overlaps
    """
    new_ranges = np.copy(gap_ranges)
    if len(gap_ranges) < 3:
        raise UserValueError(f"At least three ranges are required. Value provided was {gap_ranges}")
    if np.shape(gap_ranges)[1] != 2:
        raise UserValueError(
            f"Too many values in the ranges array. The shape of the array should "
            f"be (n,2), but we got {np.shape(gap_ranges)}"
        )

    @check_types
    def _calculate_overlaps() -> np.ndarray:
        """Calculate overlaps.

        Calculates the overlap between adjacent intervals as positive numbers and the space between intervals
        as a positive number, enforcing a minimum distance between intervals given by a buffer of data points.
        The overlap is always calculated looking to the left of an interval. For example, if two intervals
        with index locations int1=[10, 15] and int2=[16, 20] an overlap of 5 is assigned to it. If they were
        int1=[10, 15] and int2=[22, 25], an overlap of -1 (or free space of 1 index) would be assigned.
        """
        lin_ranges = np.reshape(new_ranges, (1, np.size(new_ranges))).squeeze()
        lin_ranges = np.insert(lin_ranges, 0, -1)
        lin_ranges = np.append(lin_ranges, data_length)
        overlaps = np.diff(lin_ranges)
        # Delete range lengths and account for buffers
        overlaps = 1 + buffer - np.delete(overlaps, np.arange(1, len(overlaps), 2))
        overlaps[-1] += -1
        tot_overlaps = abs(np.sum(overlaps[np.where(overlaps > 0)]))
        tot_available = np.sum(overlaps[np.where(overlaps < 0)])
        if np.sum(overlaps) > 0:
            raise UserValueError(
                f"The data points overlaps ({tot_overlaps}) is higher than the available space to "
                f"re-arrange the gaps ({tot_available}). Please use a smaller fraction of data to "
                f"remove or a longer time series"
            )
        return overlaps

    overlaps = _calculate_overlaps()

    # Check edge cases and adjust positioning of edge gaps
    ind1 = 0
    ind2 = -1
    while overlaps[ind1] > 0 or overlaps[ind2] > 0:
        if overlaps[ind1] > 0:
            new_ranges[ind1] += overlaps[ind1]
            ind1 += 1
        if overlaps[ind2] > 0:
            new_ranges[ind2] += -overlaps[ind2]
            ind2 += -1
        overlaps = _calculate_overlaps()

    # Adjust intermediate gaps that overlap by moving toe left and right depending on
    # available moves on each direction
    for idx in np.arange(1, len(overlaps) - 1):
        if overlaps[idx] > 0:
            left = -np.sum(overlaps[:idx])
            right = -np.sum(overlaps[idx + 1 :])

            if left >= right and overlaps[idx] <= left:  # Move gap to the left
                # Move intervals to the left if more and enough space available to the left
                count = 1
                while overlaps[idx] > 0:
                    if overlaps[idx - count] < 0:
                        new_ranges[idx - count : idx] += -min(overlaps[idx], -overlaps[idx - count])
                    count += 1
                    overlaps = _calculate_overlaps()

            elif left < right and overlaps[idx] <= right:  # Move gap to the right
                # Move intervals to the right if more and enough space available to the right
                count = 1
                while overlaps[idx] > 0:
                    if overlaps[idx + count] < 0:
                        new_ranges[idx : idx + count] += min(overlaps[idx], -overlaps[idx + count])
                        if idx + count == len(overlaps):
                            new_ranges[idx + count] += min(overlaps[idx], -overlaps[idx + count])

                    count += 1
                    overlaps = _calculate_overlaps()
            elif np.all(np.array([left, right]) < overlaps[idx]) and overlaps[idx] <= left + right:
                # Move adjacent gaps away from current gap
                count = 1
                while overlaps[idx] > 0:
                    # Move gaps on the right
                    if overlaps[idx + count] < 0 and idx + count <= len(overlaps):
                        new_ranges[idx : idx + count] += min(overlaps[idx], -overlaps[idx + count])
                        if idx + count == len(overlaps):
                            new_ranges[idx + count] += min(overlaps[idx], -overlaps[idx + count])

                    # Move gaps on the left
                    if overlaps[idx - count] < 0 and idx - count >= 0:
                        new_ranges[idx - count : idx] += -min(overlaps[idx], -overlaps[idx - count])

                    count += 1

                    overlaps = _calculate_overlaps()

    return new_ranges


@check_types
def _time_array(index: pd.DatetimeIndex) -> np.array:
    """Get time array.

    Convert a pandas DatetimeIndex to a time array in seconds, where t=0 is the start date of the index.

    Args:
        index: Date-time index of a pandas time series.

    Returns:
        numpy.array: Time in seconds representation of the date time index.

    """
    if index.empty:
        raise UserValueError("The date-time index is empty!")
    return (index - index[0]).view(np.int64) / 1e9


@check_types
def _make_index(
    start: Union[pd.Timedelta, pd.Timestamp, str, None] = None,
    end: Union[pd.Timedelta, pd.Timestamp, str, None] = None,
    freq: Any = pd.Timedelta(1, "m"),
) -> pd.DatetimeIndex:
    """Make datetime index.

    Method to generate a datetime index with a uniform time stamp. The time step default is 1 minute and
    a date range (total duration) defaults to 1 day, unless different parameter are provided (start, end, and sample
    frequency).

    Args:
        start: Start date-time
        end: End date-time
        freq: Frequency
            Sample frequency in time units, for example '3 h' represents 3 hours.

    Returns:
        index (pandas.DatetimeIndex): Datetime index

    Raises:
       UserValueError: If one of the inputs is wrong

    """
    # If no end and/or start dates are given, the default signal duration is set to '1 day' and end date to 'now'
    end_date = start_date = None
    if start is None and end is None:
        end = "now"
        end_date = pd.Timestamp(end)
        start_date = end_date - pd.Timedelta("1 day")
    elif start is not None and end is not None:
        end_date = pd.Timestamp(end)
        start_date = pd.Timestamp(start)
        if end_date < start_date:
            # end_date, start_date = start_date, end_date
            raise UserValueError(f"Expected start date before end date, got start={start_date}, end={end_date}")
    elif start is not None and end is None:
        start_date = pd.Timestamp(start)
        end_date = start_date + pd.Timedelta("1 day")
    elif start is None and end is not None:
        end_date = pd.Timestamp(end)
        start_date = end_date - pd.Timedelta("1 day")

    # Catch strange sample frequency inputs
    if not isinstance(freq, pd.Timedelta):
        freq = pd.Timedelta(1, "m")
        warnings.warn("Can't recognize the sample frequency, setting it to the '1 m' default.")

    index = pd.date_range(start=start_date, end=end_date, freq=freq)
    return index


@check_types
def _random_list_of_integers_that_gives_exact_sum(list_length: int, sum_output: int) -> list:
    if list_length > sum_output:
        raise UserValueError(
            f"The list length ({list_length}) is larger that the expected sum on interval lengths {sum_output}. "
            f"Use a list length lower than or equal to {sum_output}. "
            f"Note: a list length similar to the expected sum can result in zeros (i.e. no gap assigned to an interval"
        )
    if list_length < 2:
        raise UserValueError(f"The list length must be higher that 2. Value provided = {list_length}")

    # Create an list of a given size and initialize to zero
    int_list = [0] * list_length
    rng = np.random.default_rng(1975)

    # Force sum of list components to be equal to the desired sum output
    for i in range(sum_output):
        # Add 1 at random locations up to desired sum
        int_list[rng.integers(0, sum_output) % list_length] += 1

    return int_list
